##
## Filename    : chart_model.py
## Author(s)   : Michel Le Borgne
## Created     : 4/3/2010
## Revision    :
## Source      :
##
## Copyright 2009 - 2010 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
##     http:
##     mailto:
##
## Contributor(s): Nolwenn Le Meur, Geoffroy Andrieux
##
"""
Classes for representing a guarded transition model
"""
from __future__ import unicode_literals
from __future__ import print_function
from math import sqrt
from collections import defaultdict
from antlr3 import ANTLRStringStream, ANTLRFileStream, CommonTokenStream

from condexpLexer import condexpLexer
from condexpParser import condexpParser
from cadbiom.models.guard_transitions.translators.cadlangLexer import cadlangLexer
from cadbiom.models.guard_transitions.translators.cadlangParser import cadlangParser

from cadbiom import commons as cm

# number of simple nodes before we draw macros as simple nodes
MAX_SIZE_MACRO = 50
MAX_SIZE = 2000 # max number of nodes we draw

LOGGER = cm.logger()

class ChartModelException(Exception):
    """
    Exception for chart models
    """
    def __init__(self, mess):
        self.message = mess




class ChartModel(object):
    """
    Model of a chart - implements the observer pattern as subject
    observers must have an update method
    """
    def __init__(self, name, xml_namespace="http://cadbiom.genouest.org/"):
        """
        :param name: Name of the model (ex: concat of graphs uris).
        :param xml_namespace: Global namespace: Model version.
            ex: http://cadbiom.genouest.org/ (v1),
            or http://cadbiom.genouest.org/v2/ (v2)
        """
        self.name = name
        self.xml_namespace = xml_namespace   # Version of cadbiom model
        self.simple_node_dict = dict()       # for quick finding - name -> node
        self.node_dict = dict()
        self.transition_list = []            # idem
        # idem - simple node name -> list of influenced transition
        self.signal_transition_dict = dict()
        self.__root = CTopNode(name, self)
        self.constraints = "" # string of biosignal clock constraints
        self.modified = False
        self.show_macro = True
        # default value for max number of nodes for drawing
        self.max_size = MAX_SIZE
        self.__observer_list = [] # for observer pattern

    # observer pattern methods
    def attach(self, obs):
        """
        observer pattern standard attach methods
        """
        if not obs in self.__observer_list:
            self.__observer_list.append(obs)

    def detach(self, obs):
        """
        observer pattern standard detach methods
        """
        self.__observer_list.remove(obs)

    def notify(self):
        """
        observer pattern standard notify methods
        """
        for obs in self.__observer_list:
            obs.update()


    def build_from_cadlang(self, file_name, reporter):
        """
        Build a model from a .cal file of PID database
        @param file_name: str - path of .cal file
        """
        crep = reporter
        fstream = ANTLRFileStream(file_name)
        lexer = cadlangLexer(fstream)
        lexer.set_error_reporter(crep)
        parser = cadlangParser(CommonTokenStream(lexer))
        parser.set_error_reporter(crep)
        parser.cad_model(self)

    # model methods
    def draw(self, view):
        """
        @param view: chart_view
        """
        # we don't update views which are not visible
        if not view.window:
            return

        nb_snodes = len(self.simple_node_dict)
        # if model size two large don't draw
        if  nb_snodes > self.max_size:
            return

        # if model size two large, don't show macros
        if  nb_snodes > MAX_SIZE_MACRO:
            self.show_macro = False
        self.__root.draw(view)

    def get_root(self):
        """
        @return: the root of the hierarchy
        """
        if self.__root.submodel:
            return self.__root.sub_nodes[0]
        else:
            return self.__root

    def find_element(self,  m_v_c, dstyle):
        """
        @param m_v_c :coordinates of the mouse in virtual screen
        @param dstyle: drawing style (gives virtual size of fixed size nodes)

        Given the window mouse coordinates, return (node, handle, center)
        where node is the node the mouse is in,  handle the handle of the node
        the mouse is in and c are the coordinates of the node center in view.
        If no handle, handle = 0, handle are 1,2,3,4 clockwise numbered
        from upper left corner
        If no node found returns (None,0,(0,0))
        """
        return self.__root.find_element(m_v_c[0], m_v_c[1], dstyle)

    def make_submodel(self, mnode):
        """
        make a submodel from a macronode (no check) of another model
        """
        self.__root.sub_nodes = []
        self.__root.add_submodel(mnode)

    def is_submodel(self):
        """
        test if it is a submodel (for macro view)
        """
        return self.__root.submodel

    def is_modified(self):
        """
        Changes occur?
        """
        return self.modified

    def clean(self):
        """
        Clean markers
        """
        self.__root.clean()

    def clean_code(self):
        """
        Clean code attribute
        """
        self.__root.clean_code()

    def accept(self, visitor):
        """
        Visitors entrance
        """
        return visitor.visit_chart_model(self)

    def get_simple_node_names(self):
        """
        @return: list of simple nodes in model
        """
        # refresh transition dictionary
        self.signal_transition_dict = dict()
        for trans in self.transition_list:
            ids = trans.get_influencing_places()
            for ident in ids:
                try:
                    id_tr = self.signal_transition_dict[ident]
                    self.signal_transition_dict[id].append(trans)
                except:
                    self.signal_transition_dict[ident] = [trans]
        return self.simple_node_dict.keys()

    def get_matching_node_names(self, regex_obj):
        """
        @return: return node names with name matching reg. expr.
        """
        lname = []
        for name in self.simple_node_dict.keys():
            if regex_obj.match(name):
                lname.append(name)
        return lname


    def search_unmark(self):
        """
        notify observers after unmarking nodes
        """
        for key in self.simple_node_dict:
            self.simple_node_dict[key].search_mark = False
        for key in self.signal_transition_dict:
            for trans in self.signal_transition_dict[key]:
                trans.search_mark = False
        self.notify()

    def search_mark(self, lnode):
        """
        notify observers after marking nodes from lnode
        """
        # unmark first
        for key in self.simple_node_dict:
            self.simple_node_dict[key].search_mark = False
        for key in self.signal_transition_dict:
            for trans in self.signal_transition_dict[key]:
                trans.search_mark = False
        for name in lnode:
            try:
                self.simple_node_dict[name].search_mark = True
                # do not mark transition if node doesnt exist
                for trans in self.signal_transition_dict[name]:
                    trans.search_mark = True
            except:
                pass
        self.notify()

    def get_simple_node(self, name):
        """
        @param name: string - name of the node
        @return a simple node with given name
        """
        return self.simple_node_dict[name]

    def get_node(self, name):
        """
        @param name: string - name of the node
        @return: a node
        """
        return self.node_dict[name]

    def get_influenced_transition(self, node_name):
        """
        @param node_name: string
        @return: the transitions influenced by the node i.e. when
                 the node name appears in the condition
        """
        try:
            return self.signal_transition_dict[node_name]
        except:
            return []

    # method for model transformations
    def mark_as_frontier(self, node_name):
        """
        @param node_name: string
        Given the name of a simple node, add a start node and a transition
        from the start node to the simple node
        @warning: notify observers
        """
        try:
            snode = self.simple_node_dict[node_name]
            macro = snode.father
            start = macro.add_start_node(0, 0)
            macro.add_transition(start, snode)
        except KeyError:
            raise ChartModelException("Unknown simple node: "+node_name)
        self.notify()

    def __turn_into_other(self, node_name, ntype):
        """
        @param node_name: string
        @param ntype: node type - string
        turn a simple name into a permanent or input node
        The simple name must not have entering transitions
        @warning: notify observer
        """
        try:
            snode = self.simple_node_dict[node_name]
        except KeyError:
            raise ChartModelException("Unknown simple node: "+node_name)
        if len(snode.incoming_trans) != 0:
            raise ChartModelException("Incoming transition on node: "+node_name)
        macro = snode.father
        if ntype == 'perm':
            pnode = macro.add_perm_node(node_name, snode.xloc, snode.yloc)
        else:
            pnode = macro.add_input_node(node_name, snode.xloc, snode.yloc)
        for trans in snode.outgoing_tran:
            macro.add_transition(pnode, trans. ext)
        snode.remove()
        self.notify()

    def turn_into_input(self, node_name):
        """
        Turn a simple node into an input node
        """
        self.__turn_into_other(node_name, 'input')

    def turn_into_perm(self, node_name):
        """
        Turn a simple node into a perm node
        """
        self.__turn_into_other(node_name, 'perm')


class CNode(object):
    """
    Generic node for guarded transition models
    """
    wmin = 0.1
    hmin = 0.1
    depth_max = 3

    """
    Base class for  model components
    """
    def __init__(self, x_coord, y_coord, model):
        """
        The coordinates of a node are always in the space of its father.
        """
        self.model = model
        self.name = '$$'
        self.note = ""
        self.xloc = x_coord
        self.yloc = y_coord
        self.father = None
        self.selected = False
        self.search_mark = False
        self.incoming_trans = []
        self.outgoing_trans = []

    def is_top_node(self):
        """
        As it says
        """
        return False

    def is_macro(self):
        """
        As it says
        """
        return False

    def is_start(self):
        """
        As it says
        """
        return False

    def is_input(self):
        """
        As it says
        """
        return False

    def is_perm(self):
        """
        As it says
        """
        return False

    def is_trap(self):
        """
        As it says
        """
        return False

    def is_simple(self):
        """
        As it says
        """
        return False

    def set_model(self, model):
        """
        As it says
        """
        self.model = model

    def set_name(self, name):
        """
        As it says
        """
        self.name = name

#    def set_coordinates(self, x, y):
#        """
#        As it says
#        """
#        self.xloc = x
#        self.yloc = y
#        self.model.notify()

    def get_coordinates(self):
        """
        As it says
        """
        return (self.xloc, self.yloc)

    def set_layout_coordinates(self, x_coord, y_coord):
        """
        As it says
        """
        self.xloc = x_coord
        self.yloc = y_coord


    def find_element(self, mox, moy, dstyle):
        """
        The mouse coordinates must be in the same frame than nodes coordinates
        """
        pass


    def remove(self):
        """
        Remove this node from its model - assume it is not a top-node
        """
        if self.father:
            nfath = self.father
            # remove transitions using this node
            to_remove = []
            for gtr in nfath.transitions:
                if gtr[0].ori == self :
                    to_remove.append(gtr)

                elif gtr[0].ext == self :
                    to_remove.append(gtr)

            for gtr in to_remove:
                nfath.transitions.remove(gtr)
                for trans in gtr :
                    trans.remove()

            nfath.sub_nodes.remove(self)
            if self.model.simple_node_dict.has_key(self.name):
                del self.model.simple_node_dict[self.name]
            self.model.modified = True
            self.model.notify()
        else:
            pass

    def clean(self):
        """
        Abstract
        """
        pass


class CStartNode(CNode):
    """
    Start node show macro-states initialisation
    """
    def __init__(self, x_coord, y_coord, model):
        CNode.__init__(self, x_coord, y_coord, model)
        self.name = "__start__"
        self.wloc = 1.0
        self.hloc = 1.0 # for compatibility with find

    def copy(self, model = None):
        """
        As says
        """
        if not model:
            model = self.model
        node = CStartNode(self.xloc, self.yloc, model)
        return node

    def is_start(self):
        """
        As says
        """
        return True

    def is_for_origin(self):
        """
        As says
        """
        return True

    def is_for_extremity(self):
        """
        As says
        """
        return False

    def get_center_loc_coord(self, dstyle, w_ratio, h_ratio):
        """
        @param dstyle: drawing style (not used here)
        @param w_ratio,h_ratio: affinity ratios for virtual screen (not used here)

        Returns center coordinate in surrounding node - here node coordinates
        """
        return (self.xloc, self.yloc)

    def draw(self, view, xfr, yfr, wfr, hfr, depth):
        """
        depth is less than depth_max
        """
        # graphic style
        view.drawing_style.draw_start(self, xfr, yfr, wfr, hfr)

    def find_element(self, mox, moy , dstyle, w_coef, h_coef):
        """
        No handle
        """
        # bounding box in father's frame
        v_bb_size = self.accept(dstyle)
        snw = v_bb_size[0] * w_coef
        snh = v_bb_size[1] * h_coef
        hdec = v_bb_size[2] * h_coef

        # center coordinates in father's frame
        ccx = self.xloc
        ccy = self.yloc + hdec
        in_node = (mox>=ccx-snw) and (mox<=ccx+snw)
        in_node = in_node and  (moy>=ccy-snh) and (moy<=ccy+snh)
        if in_node:
            return (self, 0, (ccx, ccy), None)
        else:
            return (None, 0, (0, 0), None)

    def move(self, v_dx, v_dy, dstyle, top_node):
        """
        Move the node - mx_virt, my virt coordinates of the mouse in virtual screen frame
        click_loc is the location of the clic in node's frame
        """

        # move in reference frame (father's)
        (acx, acy) = self.father.v_affinity_coef(top_node)
        loc_dx = acx*v_dx
        loc_dy = acy*v_dy
        # translation vector in own's frame -> father's frame
        new_xloc = self.xloc + loc_dx
        new_yloc = self.yloc + loc_dy

        # TODO compute limits
        delta = 0.0
        if (new_xloc >= 0) and (new_xloc<1.0):
            self.xloc = new_xloc
        if (new_yloc > delta) and (new_yloc < 1.0):
            self.yloc = new_yloc
        self.model.modified = True
        self.model.notify()

    def intersect(self, node2, dstyle, nb_trans, w_ratio, h_ratio):
        """
        @param dstyle: drawing style
        @param w_ratio, h_ratio: affinity ratios virtual -> local frame
        """
        (xc2, yc2) = node2.get_center_loc_coord(dstyle, w_ratio, h_ratio)
        # local axes size (ellipse because of affinities)
        v_size = self.accept(dstyle)
        rlocx = v_size[0]/w_ratio
        rlocy = v_size[1]/h_ratio
        uux = xc2 - self.xloc
        uuy = yc2 - self.yloc
        norm = sqrt(uux**2 + uuy**2)
        if norm != 0:
            uux = uux / norm
            uuy = uuy / norm
        isx = self.xloc + uux * rlocx
        isy = self.yloc + uuy * rlocy
        return (isx, isy, 0.0, 1)

    def clean(self):
        pass

    def accept(self, visitor):
        """
        Generic visitor acceptor
        """
        return visitor.visit_cstart_node(self)

class CTrapNode(CStartNode):
    """
    Dead end node
    """
    def __init__(self, xcoord, ycoord, model):
        CStartNode.__init__(self, xcoord, ycoord, model)
        self.name = "__trap__"

    def is_trap(self):
        return True

    def is_start(self):
        return False

    def copy(self, model = None):
        if not model:
            model = self.model
        node = CTrapNode(self.xloc, self.yloc, model)
        return node

    def is_for_origin(self):
        return False

    def is_for_extremity(self):
        return True

    def draw(self, view, xfr, yfr, wfr, hfr, depth):
        """
        depth is less than depth_max
        """
        # graphic context
        view.drawing_style.draw_trap(self, xfr, yfr, wfr, hfr)

    def accept(self, visitor):
        """
        Standard acceptor
        """
        return visitor.visit_ctrap_node(self)

class CSimpleNode(CNode):
    """
    A simple node cannot have sub nodes
    Simple nodes have constant screen dimensions
    """
    def __init__(self, xcoord, ycoord, name, model):
        CNode.__init__(self, xcoord, ycoord, model)
        self.name = name.encode('ascii')
        self.father = None # double linkage for coordinate computations
        self.hloc = 1.0
        self.activated = False
        self.was_activated = False

    def copy(self, model = None):
        """
        As says
        """
        if not model:
            model = self.model
        node = CSimpleNode(self.xloc, self.yloc, self.name, model)
        return node

    def is_simple(self):
        return True

    def is_for_origin(self):
        """
        Can be used as transition origin
        """
        return True

    def is_for_extremity(self):
        """
        Can be used as transition extremity
        """
        return True

    def set_name(self, name):
        try:
            del self.model.simple_node_dict[self.name]
        except:
            pass
        self.name = name
        self.model.simple_node_dict[name] = self

    def get_center_loc_coord(self, dstyle, w_ratio, h_ratio):
        """
        Returns center coordinate in surrounding node
        """
        v_size = self.accept(dstyle)
        xco = self.xloc + (v_size[0] / w_ratio) / 2.0
        yco = self.yloc + (v_size[1] / h_ratio) / 2.0
        return (xco, yco)


    def draw(self, view, xfr, yfr, wfr, hfr, depth):
        """
        depth is less than depth_max
        @param xfr: x coordinate of father in view screen
        @param yfr: y coordinate of father in view screen
        @param wfr: width of father in virtual screen
        @param hfr: height of father in virtual screen
        """
        view.drawing_style.draw_simple(self, xfr, yfr, wfr, hfr)


    def find_element(self, mox, moy, dstyle, w_coef, h_coef):
        """
        Simple node - No handle
        """
        # size of the node in father's frame
        v_bb = self.accept(dstyle)
        snw = v_bb[0] * w_coef
        snh = v_bb[1] * h_coef

        # center coordinates in father's frame
        ccx = self.xloc + 0.5 * snw
        ccy = self.yloc + 0.5 * snh
        in_node = (mox >= self.xloc) and (mox <= self.xloc + snw)
        in_node = in_node and (moy >= self.yloc) and (moy <= self.yloc+snh)
        if in_node:
            return (self, 0, (ccx, ccy), None)
        else:
            return (None, 0, (0, 0), None)


    def move(self, v_dx, v_dy, v_size, top_node):
        """
        Move the node - mx_virt, my virt are virtual coordinates of the mouse
        click_loc is the location of the clic in node's frame
        """
        # move in reference frame (father's)
        (ccx, ccy) = self.father.v_affinity_coef(top_node)
        loc_dx = ccx * v_dx
        loc_dy = ccy * v_dy
        # translation vector in reference frame i.e. father's frame
        new_xloc = self.xloc + loc_dx
        new_yloc = self.yloc + loc_dy

        # check limits
        wloc = v_size[0] * ccx
        hloc = v_size[1] * ccy
        delta = v_size[2] * ccy

        move = False
        if (new_xloc >= 0) and (new_xloc+wloc < 1.0):
            self.xloc = new_xloc
            move = True
        if (new_yloc > delta) and (new_yloc+hloc < 1.0):
            self.yloc = new_yloc
            move = True
        if move:
            self.model.modified = True
            self.model.notify()


    def intersect(self, node2, dstyle, nb_trans, w_ratio, h_ratio):
        """
        Gives the the first point where to branch a transition, the gap between two arrows
        and a boolean horizontal true if arrows start from horizontal edge
        Assume wloc and hloc computed (node drawn) - coordinates in container frame

        @param node2: second node of the transition
        @param dstyle: drawing style
        @param nb_trans: number of transitions to be drawn
        @param w_ratio, h_ratio: dimentions of the screen for fixed size nodes
        """
        return intersect_simple(self, node2, dstyle, nb_trans, w_ratio, h_ratio)


    def clean(self):
        self.activated = False
        self.was_activated = False

    def accept(self, visitor):
        """
        standard visitor acceptor
        """
        return visitor.visit_csimple_node(self)

class CPermNode(CSimpleNode):
    """
    permanent node are never unactivated
    """

    def __init__(self,  xcoord, ycoord, name, model):
        CSimpleNode.__init__(self, xcoord, ycoord, name, model)

    def copy(self, model = None):
        if not model:
            model = self.model
        node = CPermNode(self.xloc, self.yloc, self.name, model)
        return node

    def is_perm(self):
        return True

    def is_simple(self):
        return False

    def is_for_extremity(self):
        return False

    def accept(self, visitor):
        return visitor.visit_cperm_node(self)

    def draw(self, view, xfr, yfr, wfr, hfr, depth):
        """
        depth is less than depth_max
        @param xfr, yfr: coordinates of father node (reference frame) in view
        @param wfr,hfr: width and height of father node in view (affinity ratios)
        """
        view.drawing_style.draw_perm(self, xfr, yfr, wfr, hfr)


class CInputNode(CSimpleNode):
    """
    An input node cannot have an in-transition
    """
    def __init__(self, xcoord, ycoord, name, model):
        CSimpleNode.__init__(self, xcoord, ycoord, name, model)
        self.father = None # double linkage for coordinate computations
        self.hloc = 1.0
        self.activated = False

    def is_input(self):
        return True

    def is_simple(self):
        return False

    def copy(self, model = None):
        if not model:
            model = self.model
        node = CInputNode(self.xloc, self.yloc, self.name, model)
        return node

    def is_for_extremity(self):
        return False

    def get_center_loc_coord(self, dstyle, w_ratio, h_ratio):
        """
        Returns center coordinate in surrounding node
        """
        return (self.xloc, self.yloc)

    def draw(self, view, xfr, yfr, wfr, hfr, depth):
        """
        special drawing for diamond
        """
        view.drawing_style.draw_input(self, xfr, yfr, wfr, hfr)

    def find_element(self, mox, moy , dstyle, w_coef, h_coef):
        """
        Simple node - No handle
        """
        # size of the box in father's frame
        v_bb = self.accept(dstyle)
        snw = v_bb[0] * w_coef
        snh = v_bb[1] * h_coef

        # center coordinates in father frame
        ccx = self.xloc + 0.5 * snw
        ccy = self.yloc + 0.5 * snh

        # mouse coordinates in self frame
        mxl = (mox - self.xloc) / snw
        myl = (moy - self.yloc) / snh
        # test
        in_node = (mox >= self.xloc) and (mox <= self.xloc+snw)
        in_node = in_node and (moy >= self.yloc) and (moy <= self.yloc+snh)

        if in_node:
            return (self, 0, (ccx, ccy), None)
        else:
            return (None, 0, (0, 0), None)

    def move(self, v_dx, v_dy, v_size, top_node):
        """
        Move the node - mx_virt, my virt are virtual coordinates of the mouse
        click_loc is the location of the clic in node's frame
        """
        # move in reference frame (father's)
        (ccx, ccy) = self.father.v_affinity_coef(top_node)
        loc_dx = ccx * v_dx
        loc_dy = ccy * v_dy
        # translation vector in reference frame i.e. father's frame
        new_xloc = self.xloc + loc_dx
        new_yloc = self.yloc + loc_dy
        # check limits
        wloc = (v_size[0] * ccx)/2.0
        hloc = (v_size[1] * ccy)/2.0
        hlabel = v_size[2] * ccy

        move = False
        if (new_xloc >= wloc) and (new_xloc+wloc < 1.0):
            self.xloc = new_xloc
            move = True
        if (new_yloc - hloc> hlabel) and (new_yloc+hloc < 1.0):
            self.yloc = new_yloc
            move = True
        if move:
            self.model.modified = True
            self.model.notify()


    def intersect(self, node2, dstyle, nb_trans, w_ratio, h_ratio):
        """
        An input node can be the origin of some transition to a node.
        One transition by node.
        """
        # local coordinates of node2 center
        (xc2, yc2) = node2.get_center_loc_coord(dstyle, w_ratio, h_ratio)
        # simple nodes have constant screen dimensions - convert to local ones
        v_size = self.accept(dstyle)
        wloc = (v_size[0] / w_ratio) / 2.0
        hloc = (v_size[1] / h_ratio) / 2.0
        # coordinates of a diamond are center coordinates
        uux = xc2 - self.xloc
        uuy = yc2 - self.yloc
        if uux != 0.0:
            slope = uuy / uux
            if slope < -1:
                if uux < 0.0:
                    return (self.xloc, self.yloc + hloc, 0, True) # upper corner
                else:
                    return (self.xloc, self.yloc - hloc, 0, True) # low corner
            elif slope >= -1 and slope <= 1:
                if uux < 0.0:
                    return (self.xloc - wloc, self.yloc, 0, True) # left corner
                else:
                    return (self.xloc + wloc, self.yloc, 0, True) # right corner
            else:
                if uux < 0.0:
                    return (self.xloc, self.yloc - hloc, 0, True) # low corner
                else:
                    return (self.xloc, self.yloc + hloc, 0, True) # upper corner
        else:
            if uuy >= 0:
                return (self.xloc, self.yloc + hloc, 0, True) # upper corner
            else:
                return (self.xloc, self.yloc - hloc, 0, True) # low corner

    def accept(self, visitor):
        return visitor.visit_cinput_node(self)

class CMacroNode(CSimpleNode):
    """
    Main building block for charts
    """
    def __init__(self, xcoord, ycoord, width, height, name, model):
        CSimpleNode.__init__(self, xcoord, ycoord, name, model)
        self.count = 0 # for start and trap nodes naming
        self.wloc = width
        self.hloc = height
        self.sub_nodes = []
        # list<list<CTransitions>> Sublists: transitions with common extremities
        # self.transitions = []
        # See the property 'transitions' to find a wrapper for old code
        self.new_transitions = defaultdict(list)

    @property
    def transitions(self):
        """Compatibility code
        .. note:: The old API uses this attribute as a list<list<CTransitions>>.
            Sublists are transitions with common extremities.

        :return: An iterator over the values of self.new_transitions.
            Similar to: <list <list <CTransitions>>>
        :rtype: <dictionary-valueiterator>
        """
        return self.new_transitions.itervalues()

    @transitions.setter
    def transitions(self, _):
        """Block further modifications of the old attribute transitions
        .. note:: Please modify self.new_transitions instead.
        """
        raise Exception("NOT AUTHORIZED! Please modify new_transitions instead")

    def _find_in_subnodes(self, node):
        """
        Find a node in the list of sub nodes.
        assume two nodes have different coordinates
        @param node: node to be found
        """
        for snode in self.sub_nodes:
            if snode.xloc == node.xloc and snode.yloc == node.yloc:
                return snode
        return None

    def copy(self, model = None):
        """
        Duplicate a macronode - performs a deep copy
        """
        if not model:
            model = self.model
        node_c = CMacroNode(self.xloc, self.yloc, self.wloc, self.hloc,
                            self.name, model)
        # copy subnodes
        for snode in self.sub_nodes:
            snc = snode.copy(model)
            node_c.sub_nodes.append(snc)
            snc.father = node_c
        # copy transitions
        for gtr in self.transitions:
            gtr_c = []
            for trans in gtr:
                origin = trans.ori
                origin_c = node_c._find_in_subnodes(origin)
                extremity = trans.ext
                extremity_c = node_c._find_in_subnodes(extremity)
                trc = CTransition(origin_c, extremity_c)
                trc.macro_node = node_c
                trc.event = trans.event
                trc.condition = trans.condition
                trc.action = trans.action
                gtr_c.append(trc)
            node_c.transitions.append(gtr_c)
        return node_c

    def set_model(self, model):
        """
        Set the model for a subtree of nodes - used for copy/paste from one model to another
        """
        for snode in self.sub_nodes:
            snode.set_model(model)
        self.model = model
        self.model.modified = True

    def set_name(self, name):
        self.name = name

    def is_macro(self):
        return True

    def is_simple(self):
        return False

    def is_for_origin(self):
        """
        Legitimate transition origin
        """
        return True

    def is_for_extremity(self):
        """
        Legitimate transition extremity
        """
        return True

    def get_center_loc_coord(self, dstyle, w_ratio, h_ratio):
        """
        Returns center coordinate in surrounding node (father)
        """
        if self.model.show_macro:
            xcc = self.xloc + self.wloc / 2.0
            ycc = self.yloc + self.hloc / 2.0
        else:
            v_size = self.accept(dstyle)
            xcc = self.xloc + (v_size[0] / w_ratio) / 2.0
            ycc = self.yloc + (v_size[1] / h_ratio) / 2.0
        return (xcc, ycc)

    def virt_to_self_frame(self, xcoord, ycoord, s_width, s_height, top_node):
        """
        xcoord,ycoord must be virtual coordinates
        result is the coordinates of (x,y) in self frame
        @param xcoord, ycoord: coordinate in virtual window (screen -> 1.0 x 1.0 window)
        @param s_width, s_height: screen dimensions
        @param  top_node: root of the sub_model
        """

        if self != top_node:
            (xfath, yfath) = self.father.virt_to_self_frame(xcoord,
                                                            ycoord,
                                                            s_width, s_height,
                                                            top_node)
            xloc = (xfath - self.xloc) / self.wloc
            yloc = (yfath - self.yloc) / self.hloc
            return (xloc, yloc)
        else: # top node
            return(xcoord, ycoord)

    def self_to_virtual_frame(self, xcoord, ycoord, top_node):
        """
        Coordinate change
        """
        if self != top_node:
            xfath = self.xloc + xcoord * self.wloc
            yfath = self.yloc + ycoord * self.hloc
            return self.father.self_to_virtual_frame(xfath, yfath, top_node)
        else:
            return (xcoord, ycoord)

    def v_affinity_coef(self, top_node):
        """
        Affinity ratios cw,ch: local_horizontal_length = screen_horizontal_length * cw
        Similar relation for vertical length.
        @param  top_node: root of the sub_model
        """
        if(self!= top_node):
            (acw, ach) = self.father.v_affinity_coef(top_node)
            return (acw / self.wloc, ach / self.hloc)
        else:
            return (1.0, 1.0)

    def add_macro_subnode(self, name, xcoord, ycoord, width, height):
        """
        Add a macro node as subnode with dimensions
        """
        node = CMacroNode(xcoord, ycoord, width, height, name, self.model)
        self.model.node_dict[name] = node
        node.father = self
        self.sub_nodes.append(node)
        self.model.modified = True
        self.model.notify()
        return node

    def add_simple_node(self, name,  xcoord, ycoord):
        """
        Add a simple node
        """
        node = CSimpleNode(xcoord, ycoord, name, self.model)
        self.model.node_dict[name] = node
        node.father = self
        self.sub_nodes.append(node)
        self.model.simple_node_dict[name] = node
        self.model.modified = True
        self.model.notify()
        return node

    def add_input_node(self, name,  xcoord, ycoord):
        """
        Add input node
        """
        node = CInputNode(xcoord, ycoord, name, self.model)
        self.model.node_dict[name] = node
        node.father = self
        self.sub_nodes.append(node)
        self.model.modified = True
        self.model.notify()
        return node

    def add_copy(self, node):
        """
        add a node of the same type
        """
        nnode = node.copy(self.model)
        self.model.node_dict[nnode.name] = nnode
        nnode.father = self
        if nnode.is_simple():
            self.model.simple_node_dict[nnode.name] = nnode
        self.sub_nodes.append(nnode)
        self.model.modified = True
        self.model.notify()
        return nnode

    def add_transition(self, ori, ext):
        """Add a transition to the model

        .. note:: Reflexive transitions are not authorized.
            Duplications of transitions are not authorized.
            BUT! Not exception is raised in these cases. Have fun <3

        :param ori: a simple node
        :param ext: a simple node
        :return: A new transition or None if the couple of nodes was not valid.
        :rtype: <CTransition> or None
        """
        # transition between a node and itself are forbidden
        if ori == ext:
            LOGGER.warning(
                "Reflexive transition: " + ori.name + ' -> ' + ext.name +
                " - This transition will not be taken into account.\n"
                "Please review your model or use a PermanentNode instead."
            )
            return

        # Search the current couple of nodes in all the transitions
        # self.new_transitions is a defaultdict with frozensets as keys
        # and lists as values. Each value has at most 2 transitions
        # (one in each direction). Each value is called "transitions group"
        # later in the code.
        nodes_couple = frozenset((ori.name, ext.name))
        transitions_group = self.new_transitions[nodes_couple]
        if len(transitions_group) == 0:
            # New transition: the couple was nod found before
            return self.build_new_transition_to_nodes(
                transitions_group, ori, ext
            )
        elif len(transitions_group) == 1:
            # Duplication of a transition ?
            trans = transitions_group[0]
            if not ((trans.ori == ori) and (trans.ext == ext)):
                # Incomplete transition: add the other one
                return self.build_new_transition_to_nodes(
                    transitions_group, ori, ext
                )
            else:
                LOGGER.warning(
                    "Duplicated transition: " + ori.name + ' -> ' + ext.name +
                    "\nYou should review your model. Only the first transition "
                    "will be taken into account."
                )
                return
        else:
            LOGGER.error("More than one transition for a couple of nodes")
            LOGGER.error("Current transition:" + ori.name + ' -> ' + ext.name)
            raise Exception(
                "ERROR: More than one transition for a couple of nodes"
            )

        LOGGER.error("Hi! You should never have reached this part.")
        LOGGER.error("Current transition:" + ori.name + ' -> ' + ext.name)
        raise Exception("Hi! You should never have reached this part.")


    def build_new_transition_to_nodes(self, transitions_group, ori, ext):
        """Handle a new transition: build and attach it to the model

        :param arg1: List of transitions that concern the given couple of nodes
            The list should not exceed 2 elements (see add_transition())
        :param arg2: Node
        :param arg3: Node
        :type arg1: <list <CTransition>>
        :type arg2: <Node>
        :type arg3: <Node>
        """
        # Build new transition
        new_transition = CTransition(ori, ext)
        new_transition.macro_node = self
        # Append the new transition to the internal structure that groups
        # couple of transitions with same nodes
        transitions_group.append(new_transition)
        # Append new transition to the concerned ori and ext nodes
        new_transition.ori.outgoing_trans.append(new_transition)
        new_transition.ext.incoming_trans.append(new_transition)
        # Append new transition to the model
        self.model.transition_list.append(new_transition)
        self.model.modified = True
        self.model.notify()
        return new_transition

    def add_start_node(self, xcoord, ycoord, name = None):
        """
        Add a start node
        """
        nnode = CStartNode(xcoord, ycoord, self.model)
        self.model.node_dict[name] = nnode
        nnode.father = self
        if name:
            nnode.name = name
        else:
            # add a number to have different graphic start nodes
            nnode.name = nnode.name + "%s" % self.count
        self.count = self.count + 1
        self.sub_nodes.append(nnode)
        self.model.modified = True
        self.model.notify()
        return nnode

    def add_trap_node(self, xcoord, ycoord, name = None):
        """
        Add a trap node
        """
        nnode = CTrapNode(xcoord, ycoord, self.model)
        self.model.node_dict[name] = nnode
        nnode.father = self
        if name:
            nnode.name = name
        else:
            # add a number to have different graphic trap nodes
            nnode.name = nnode.name + "%s" % self.count
        self.count = self.count + 1
        self.sub_nodes.append(nnode)
        self.model.modified = True
        self.model.notify()
        return nnode

    def add_perm_node(self, name, xcoord, ycoord):
        """
        Add a perm node
        """
        nnode = CPermNode(xcoord, ycoord, name, self.model)
        self.model.node_dict[name] = nnode
        nnode.father = self
        self.sub_nodes.append(nnode)
        self.model.modified = True
        self.model.notify()
        return nnode



    def draw(self, view, xfr, yfr, wfr, hfr, depth):
        """
        depth is less than depth_max
        @param view: drawing area
        @param xfr: x coordinate of father in virtual screen
        @param yfr: y -
        @param wfr: father's width in virtual screen
        @param hfr: father's height in virtual
        @param depth: depth of the node
        """
        dstyle = view.drawing_style
        # draw node
        dstyle.draw_macro(self, xfr, yfr, wfr, hfr)
        # draw sub graph
        if depth < CMacroNode.depth_max and self.model.show_macro:
            xxr = xfr + self.xloc * wfr
            yyr = yfr + self.yloc * hfr
            w_ratio = wfr * self.wloc
            h_ratio = hfr * self.hloc
            # edges
            for tgr in self.transitions:
                dstyle.draw_transition_group(tgr, xxr, yyr, w_ratio, h_ratio)
            # nodes
            for snode in self.sub_nodes:
                snode.draw(view, xxr, yyr, w_ratio, h_ratio, depth + 1)


    def move(self, v_dx, v_dy, v_size, top_node):
        """
        Move the node - mx_virt, my virt are virtual coordinates of the mouse
        click_loc (pair) is the location of the clic in node's frame
        """
        # don't try to move a root node
        if self == top_node:
            return
        # move in reference frame (father's)
        (ccx, ccy) = self.father.v_affinity_coef(top_node)
        loc_dx = ccx * v_dx
        loc_dy = ccy * v_dy
        # translation vector in reference frame i.e. father's frame
        new_xloc = self.xloc + loc_dx
        new_yloc = self.yloc + loc_dy
        # check limits
        wloc = v_size[0] * ccx
        hloc = v_size[1] * ccy
        delta = v_size[2] * ccy

        if self.model.show_macro:
            wloc = self.wloc
            hloc = self.hloc
        else:
            wloc = v_size[0] * ccx
            hloc = v_size[1] * ccy
        move = False
        if (new_xloc >= 0) and (new_xloc + wloc < 1.0):
            self.xloc = new_xloc
            move = True
        if (new_yloc > delta) and (new_yloc + hloc < 1.0):
            self.yloc = new_yloc
            move = True
        if move:
            self.model.modified = True
            self.model.notify()


    def find_element(self, mox, moy, dstyle, w_coef, h_coef):
        """
        @param mox, moy: mouse coordinate in container frame
        @param dstyle: drawing style
        @param w_coef, h_coef: affinity ratios for view -> container frame
        """
        # width and height of the node in container frame
        if self.model.show_macro:
            wloc = self.wloc
            hloc = self.hloc
        else:
            bb_size = self.accept(dstyle)
            wloc = bb_size[0] * w_coef
            hloc = bb_size[1] * h_coef
        in_node = (mox >= self.xloc) and (mox <= self.xloc + wloc)
        in_node = in_node and (moy >= self.yloc) and (moy <= self.yloc + hloc)
        if in_node:
            if not self.model.show_macro:
                ccx = self.xloc + 0.5 * wloc
                ccy = self.yloc + 0.5 * hloc
                return (self, 0, (ccx, ccy), None)

            # change coordinates and affinity ratio for self frame
            w_coefloc = w_coef / wloc
            h_coefloc = h_coef / hloc
            mxloc = (mox - self.xloc) / wloc
            myloc = (moy - self.yloc) / hloc
            # search sub node
            for snode in self.sub_nodes:
                (nnn, hhh, ccc, ttt) = snode.find_element(mxloc, myloc, dstyle,
                                               w_coefloc, h_coefloc)
                if nnn:
                    # n center coordinates in container frame
                    ccx = self.xloc + ccc[0] * self.wloc
                    ccy = self.yloc + ccc[1] * self.hloc
                    return (nnn, hhh, (ccx, ccy), ttt)


            # self center coordinate in container frame
            ccx = self.xloc + 0.5 * self.wloc
            ccy = self.yloc + 0.5 * self.hloc

            # not in a subnode, find a transition at this level
            trans = self.find_transition(mxloc, myloc, dstyle,
                                         w_coefloc, h_coefloc)
            if trans:
                return (self, 0, (ccx, ccy), trans)

            # not in a subnode or transition- find handler
            v_size = self.accept(dstyle)
            ddx = v_size[3]
            ddy = v_size[4]

            if (mox < self.xloc + ddx) and (moy < self.yloc + ddy):
                return (self, 1, (ccx, ccy), None)
            if (mox > self.xloc + wloc - ddx) and (moy < self.yloc + ddy):
                return (self, 2, (ccx, ccy), None)
            cond = (mox > self.xloc + wloc - ddx)
            cond = cond and (moy > self.yloc + hloc - ddy)
            if cond:
                return (self, 3, (ccx, ccy), None)
            if (mox < self.xloc + ddx) and (moy > self.yloc + hloc - ddy):
                return (self, 4, (ccx, ccy), None)
            return (self, 0, (ccx, ccy), None)
        else:
            return (None, 0, (0, 0), None)

    def find_transition(self, mox, moy, dstyle, w_coef, h_coef):
        """
        Look for transitions pointed by mouse
        No recursive search - transitions are in current node

        """
        for gtr in self.transitions:
            for trans in gtr:
                if trans.is_me(mox, moy, dstyle, w_coef, h_coef):
                    return trans
        return None


    def resize(self, mx_virt, my_virt, handle, screen_w, screen_h, top_node):
        """
        @param  mx_virt, my_virt:  mouse coordinates in virtual window
        @param handle: as determined by find_node
        @param screen_w, screen_h: size of the screen in pixels
        @param top_node: root node of the sub model
        """
        # don't try to resize a root node
        if self == top_node:
            return

        (mx_loc, my_loc) = self.father.virt_to_self_frame(mx_virt, my_virt,
                                                          screen_w, screen_h,
                                                          top_node)
        if handle == 1: # top left corner
            change = False
            new_xloc = mx_loc
            new_wloc = self.xloc + self.wloc - mx_loc
            if new_xloc > 0 and new_wloc > CNode.wmin :
                self.xloc = new_xloc
                self.wloc = new_wloc
                change = True
            new_yloc = my_loc
            new_hloc = self.yloc + self.hloc - my_loc
            if new_yloc > 0 and new_hloc > CNode.hmin :
                self.yloc = new_yloc
                self.hloc = new_hloc
                change = True
            if change:
                self.model.modified = True
                self.model.notify()

        elif handle == 2: # top right corner
            change = False
            new_wloc = mx_loc - self.xloc
            if self.xloc + new_wloc < 1.0 and new_wloc > CNode.wmin :
                self.wloc = new_wloc
                change = True
            new_yloc = my_loc
            new_hloc = self.yloc + self.hloc - my_loc
            if new_yloc > 0 and new_hloc > CNode.hmin :
                self.yloc = new_yloc
                self.hloc = new_hloc
                change = True
            if change:
                self.model.modified = True
                self.model.notify()

        elif handle == 3: # bottom right corner
            change = False
            new_wloc = mx_loc - self.xloc
            if self.xloc + new_wloc < 1.0 and new_wloc > CNode.wmin :
                self.wloc = new_wloc
                change = True
            new_hloc = my_loc - self.yloc
            if self.yloc+new_hloc < 1.0 and new_hloc > CNode.hmin :
                self.hloc = new_hloc
                change = True
            if change:
                self.model.modified = True
                self.model.notify()

        elif handle == 4: # bottom left corner
            change = False
            new_xloc = mx_loc
            new_wloc = self.xloc + self.wloc - mx_loc
            if new_xloc > 0 and new_wloc > CNode.wmin :
                self.xloc = new_xloc
                self.wloc = new_wloc
                change = True
            new_hloc = my_loc - self.yloc
            if self.yloc + new_hloc < 1.0 and new_hloc > CNode.hmin :
                self.hloc = new_hloc
                change = True
            if change:
                self.model.modified = True
                self.model.notify()
        else:
            pass

    def intersect(self, node2, dstyle, nb_trans, w_ratio, h_ratio):
        """
        Gives the the first point where to branch a transition, the gap between two arrows
        and a boolean horizontal true if arrows start from horizontal edge
        @param node2: second node of the transition
        @param dstyle: drawing style
        @param nb_trans: number of transitions to be drawn
        @param w_ratio,h_ratio: dimentions of the screen for fixed size nodes
        """
        if not self.model.show_macro:
            return intersect_simple(self, node2, dstyle, nb_trans,
                                    w_ratio, h_ratio)
        (xc2, yc2) = node2.get_center_loc_coord(dstyle, w_ratio, h_ratio)
        uux = xc2 - (self.xloc + self.wloc / 2.0)
        uuy = yc2 - (self.yloc + self.hloc / 2.0)
        wloc = self.wloc
        hloc = self.hloc
        limit_slope = hloc/wloc

        if uux != 0.0:
            slope = uuy / uux
            if slope <= -limit_slope:
                if uux > 0:
                    side = 1
                else:
                    side = 3
            elif slope >= limit_slope:
                if uux < 0:
                    side = 1
                else:
                    side = 3
            else: # -limit_slope < slope < limit_slope
                if uux > 0:
                    side = 2
                else:
                    side = 4
        else:
            if uuy >= 0 :
                side = 3
            else:
                side = 1

        if side == 1:
            horizontal = True
            gap = wloc / (nb_trans+1)
            six = self.xloc + gap
            siy = self.yloc
            return (six, siy, gap, horizontal)
        if side == 3:
            horizontal = True
            gap = wloc / (nb_trans+1)
            six = self.xloc + gap
            siy = self.yloc + hloc
            return (six, siy, gap, horizontal)
        if side == 2:
            horizontal = False
            gap = hloc / (nb_trans + 1)
            six = self.xloc + wloc
            siy = self.yloc + gap
            return (six, siy, gap, horizontal)
        if side == 4:
            horizontal = False
            gap = hloc / (nb_trans+1)
            six = self.xloc
            siy = self.yloc + gap
            return (six, siy, gap, horizontal)



    def clean(self):
        self.activated = False
        for node in self.sub_nodes:
            node.clean()
        for gtr in self.transitions:
            for trans in gtr:
                trans.clean()

    def clean_code(self):
        """
        clean the code
        """
        for ltr in self.transitions:
            for trans in ltr:
                trans.clock = None
        for node in self.sub_nodes:
            if node.is_macro():
                node.clean_code()

    def accept(self, visitor):
        """
        Standard visitor acceptor
        """
        return visitor.visit_cmacro_node(self)



class CTopNode(CMacroNode):
    """
    A virtual macronode on top of the hierarchy. A model can be a list of hierarchy.
    This artificial node group all the hierarchy. From a graphical point of view, it
    represents the virtual drawing window.
    """
    def __init__(self, name, model):
        CMacroNode.__init__(self, 0.0, 0.0, 1.0, 1.0, name, model)
        self.submodel = False
        self.env_node = None  # for environment interaction in tests
        #nb no father node

    def copy(self):
        """
        Duplicate a topnode - performs a deep copy - reduce dimensions
        The top node is changed into a macro node - the env_node (if any) is not copied.
        """
        node_c = CMacroNode(self.xloc, self.yloc, 0.5, 0.30,
                            self.name, self.model)
        # copy subnodes
        for snode in self.sub_nodes:
            snc = snode.copy()
            node_c.sub_nodes.append(snc)
            snc.father = node_c
        # copy transitions
        for gtr in self.transitions:
            gtr_c = []
            for trans in gtr:
                origin = trans.ori
                origin_c = node_c._find_in_subnodes(origin)
                extremity = trans.ext
                extremity_c = node_c._find_in_subnodes(extremity)
                trc = CTransition(origin_c, extremity_c)
                trc.macro_node = node_c
                trc.event = trans.event
                trc.condition = trans.condition
                trc.action = trans.action
                gtr_c.append(trc)
            node_c.transitions.append(gtr_c)
        return node_c

    def is_for_origin(self):
        return False

    def is_for_extremity(self):
        return False

    def is_top_node(self):
        return True

    def draw(self, view):
        # the TOP node is virtual - so no drawing
        # if there is only one subnode which is a macro - draw it full screen
        dstyle = view.drawing_style
        if self.submodel:
            mno = self.sub_nodes[0]
            w_ratio = 1.0 / mno.wloc
            h_ratio = 1.0 / mno.hloc
            xxr = - mno.xloc * w_ratio
            yyr = - mno.yloc * h_ratio
            mno.draw(view, xxr, yyr, w_ratio, h_ratio, 0)
            return
        # top of a full model
        # transitions
        for tgr in self.transitions:
            if self.submodel:
                dstyle.draw_transition_group(tgr, 0.0, 0.0, 1.0, 1.0,
                                             self.sub_nodes[0])
            else:
                dstyle.draw_transition_group(tgr, 0.0, 0.0, 1.0, 1.0)
        # nodes
        for snode in self.sub_nodes:
            snode.draw(view, 0.0, 0.0, 1.0, 1.0, 1)


    def find_element(self, mox, moy, dstyle):
        """
        @param mox,moy :coordinates of the mouse in local frame (own frame for top node)
        @param dstyle: drawing style used in the view

        Given the window mouse coordinates, return (node, handle, center) where node is the
        node the mouse is in,  handle the handle of the node the mouse is in and c are the coordinates
        of the node center in view.
        If no handle, handle = 0, handle are 1,2,3,4 clockwise numbered from upper left corner
        If no node found returns (None,0,(0,0))
        """
        # if model size two large don't draw => don't search
        nb_snodes = len(self.model.simple_node_dict)
        if  nb_snodes > self.model.max_size:
            return (self, 0, (0, 0), None) # no handle for top node

        if self.submodel:
            #TODO check new version with virtual screen
            mno = self.sub_nodes[0]
            # mouse coordinates in father frame
            mox = mox*mno.wloc + mno.xloc
            moy = moy*mno.hloc + mno.yloc
            # top node coefs in father frame
            w_coef = mno.wloc
            h_coef = mno.hloc
            return mno.find_element(mox, moy, dstyle, w_coef, h_coef)
        w_coef = 1.0
        h_coef = 1.0
        for snode in self.sub_nodes:
            (nnn, hhh, ccc, ttt) = snode.find_element(mox, moy, dstyle,
                                                       w_coef, h_coef)
            if nnn:
                # center in  screen
                ccx = ccc[0] / w_coef
                ccy = ccc[1] / h_coef
                return (nnn, hhh, (ccx, ccy), ttt)
        # no subnode found: try to find a transition at top level
        for trg in self.transitions:
            for trans in trg:
                if trans.is_me(mox, moy, dstyle, w_coef, h_coef):
                    ccx = 0.5 / w_coef
                    ccy = 0.5 / h_coef
                    return (self, 0, (ccx, ccy), trans)
        # no subnode and no transition found
        return (self, 0, (0, 0), None) # no handle for top node


#    def aff_coefs(self, top_node):
#        """
#        Affinity coefficients
#        """
#        return (1.0, 1.0)

    def aff_coef(self, swidth, sheight, top_node):
        """
        Affinity coefficients
        """
        return (1.0 / swidth, 1.0 / sheight)

    def move(self, v_dx, v_dy, v_size, top_node):
        pass # no move for top node


    def add_submodel(self, mnode):
        """
        add a submodel (subtree) with root a macro node (no check performed)
        """
        self.submodel = True
        self.sub_nodes.append(mnode)
        self.model.modified = True

    def accept(self, visitor):
        return visitor.visit_ctop_node(self)

class CTransition(object):
    """
    A guarded transition object
    """
    def __init__(self, origin, extremity):
        """
        @param origin : CNode
        @param extremity : CNode
        """
        self.macro_node = None
        self.ori = origin
        self.ext = extremity
        self.name = ""
        self.event = ""
        self.condition = ""
        self.action = ""
        self.selected = False
        self.search_mark = False
        self.activated = False
        self.fact_ids = [] #list of associated facts id
        self.note = ""
        # for compiler (avoid redondant clause generation)
        self.clock = None
        self.ori_coord = 0.0 # to be set by layout
        self.ext_coord = 0.0

    def set_event(self, event):
        """
        @param event: string
        """
        self.event = event
        self.ori.model.modified = True

    def set_condition(self, cond):
        """
        @param cond : string
        """
        self.condition = cond
        self.ori.model.modified = True

    def set_action(self, act):
        """
        @param act: string
        """
        self.action = act
        self.ori.model.modified = True

    def set_name(self, name):
        """
        @param name: string
        """
        self.name = name
        self.ori.model.modified = True

    def set_note(self, note):
        """
        @param note: string
        """
        self.note = note

    def get_key(self):
        """
        key for storing the transition in dictionaries
        """
        key = self.ori.name + self.ext.name + self.event
        key = key + self.condition + self.action
        return key

    def clean(self):
        """
        Unmark
        """
        self.activated = False

    def get_influencing_places(self):
        """
        find places that influence the condition
        """
        if len(self.condition) == 0:
            return []
        text_c = self.condition + "$"
        reader = ANTLRStringStream(text_c)
        lexer = condexpLexer(reader)
        parser = condexpParser(CommonTokenStream(lexer))
        try:
            influ = parser.sig_bool()
        except:
            return []
        return influ

    def is_me(self, mox, moy, dstyle, w_coef, h_coef):
        """
        Tell if mouse position is closed to transition
        @param mox, moy: mouse coordinates in container coord (local coordinates)
        """
        v_size = self.accept(dstyle)
        dist_max = max(v_size[0], v_size[1])
        # extremities
        (x_or, y_or) = self.ori_coord
        (x_tg, y_tg) = self.ext_coord

        # distance from mouse cursor
        llx = x_tg - x_or
        lly = y_tg - y_or
        norm = sqrt(llx * llx + lly * lly)
        if norm == 0:
            return # must not happen
        llx = llx / norm
        lly = lly / norm
        xxx = mox - x_or
        yyy = moy - y_or
        uuu = llx * xxx + lly * yyy
        dist = (xxx - uuu * llx) * (xxx - uuu * llx)
        dist = dist + (yyy - uuu * lly) * (yyy - uuu * lly)
        dist = sqrt(dist)

        if dist < dist_max:
            # are we in the bounds
            ddx = abs(x_or-x_tg)
            ddy = abs(y_or-y_tg)
            if ddx > 0.05:
                condx =  (abs(mox - x_or) <= ddx)  and (abs(mox - x_tg) <= ddx)
            else:
                condx = True
            if ddy > 0.05:
                condy = (abs(moy - y_or) <= ddy) and (abs(moy - y_tg) <= ddy)
            else:
                condy = True
            return condx and condy
        else:
            return False

    def remove(self):
        """
        remove the transition from its macro node
        """
        # list<list<CTransitions>> Sublists: transitions with common extremities
        # defaultdict(<type 'list'>, {frozenset(['A', 'D']): [D -> A, C:, E:],
        temp_transitions = self.macro_node.new_transitions
        for nodes, transitions in self.macro_node.new_transitions.items():
            if self in transitions:
                if len(transitions) == 1:
                    temp_transitions.pop(nodes)
                    continue
                # Multiple transitions: remove only the current one
                # Is it really happen?
                print("Remove only the current transition")
                transitions.remove(self)

        self.macro_node.new_transitions = temp_transitions
        self.ori.outgoing_trans.remove(self)
        self.ext.incoming_trans.remove(self)
        self.ori.model.transition_list.remove(self)
        self.ori.model.modified = True

    def accept(self, visitor):
        """
        standard accept visitor
        """
        return visitor.visit_ctransition(self)

    def __repr__(self):
        return "{} -> {}, C:{}, E:{}".format(
            self.ori.name, self.ext.name,
            self.condition, self.event)


# helper function

def intersect_simple(node1, node2, dstyle, nb_trans, w_ratio, h_ratio):
    """
    Intersection of a node with a transition
    """
    # simple nodes have constant screen dimensions - convert to local ones
    v_size = node1.accept(dstyle)
    wloc_snode = v_size[0] / w_ratio
    hloc_snode = v_size[1] / h_ratio

    (xc2, yc2) = node2.get_center_loc_coord(dstyle, w_ratio, h_ratio)
    uux = xc2 - (node1.xloc + wloc_snode/2.0)
    uuy = yc2 - (node1.yloc + hloc_snode/2.0)
    if uux != 0.0:
        slope = uuy / uux
        u_slope = v_size[1] / v_size[0]
        if slope <= -u_slope:
            if uux > 0:
                side = 1
            else:
                side = 3
        elif slope >= u_slope:
            if uux < 0:
                side = 1
            else:
                side = 3
        else: # -U_SLOPE < slope < U_SLOPE
            if uux > 0:
                side = 2
            else:
                side = 4
    else:
        if uuy >= 0 :
            side = 3
        else:
            side = 1

    if side == 1: # hight horizontal
        horizontal = True
        if nb_trans ==  1:
            six = node1.xloc + 0.5 * wloc_snode
            gap = 0.0
        elif nb_trans == 2:
            gap = wloc_snode * 0.6
            six = node1.xloc + 0.2 * wloc_snode
        siy = node1.yloc
        return (six, siy, gap, horizontal)
    if side == 3: # low horizontal
        horizontal = True
        if nb_trans ==  1:
            six = node1.xloc + 0.5 * wloc_snode
            gap = 0.0
        elif nb_trans == 2:
            gap = wloc_snode * 0.6
            six = node1.xloc + 0.2 * wloc_snode
        siy = node1.yloc + hloc_snode
        return (six, siy, gap, horizontal)
    if side == 2: # right vertical
        horizontal = False
        if nb_trans ==  1:
            siy = node1.yloc + 0.5 * hloc_snode
            gap = 0.0
        elif nb_trans == 2:
            gap = hloc_snode * 0.6
            siy = node1.yloc + 0.2 * hloc_snode
        six = node1.xloc + wloc_snode
        return (six, siy, gap, horizontal)
    if side == 4: # left vertical
        horizontal = False
        if nb_trans ==  1:
            siy = node1.yloc + 0.5 * hloc_snode
            gap = 0.0
        elif nb_trans == 2:
            gap = hloc_snode * 0.6
            siy = node1.yloc + 0.2 * hloc_snode
        six = node1.xloc
        return (six, siy, gap, horizontal)







## Filename    : gt_visitors
## Author(s)   : Michel Le Borgne
## Created     : 01/2012
## Revision    :
## Source      :
##
## Copyright 2010 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##
"""
Utilities for compiling conditions, event and constraints
"""
from antlr3 import ANTLRStringStream, CommonTokenStream, RecognitionException

from cadbiom.models.biosignal.translators.sigexpr_lexer import sigexpr_lexer
from cadbiom.models.biosignal.translators.sigexpr_compiler import sigexpr_compiler


# GT2MCSysVisitor compile a MC model from a guarded transition model
# mangling constants
ENVAC = "_envac"
ENVINAC = "_envinac"
ACT = "_activ"
INACT = "_inac"
UPDAT = "_update"


def compile_cond(text, symb_table, error_reporter, deep=-1, message=""):
    """Compile a condition expression to a tree representation

    @param text: string to be compiled
    @param symb_table: symbol table of the compiler
    @param error_reporter: use a CompilReporter for full information!!
    @param deep: optional - if >0 the compiler doesn't accept states
                declared inside macro
    @return: a tree representing the expression
    """
    text_c = text + " $"
    reader = ANTLRStringStream(text_c)
    lexer = sigexpr_lexer(reader)
    parser = sigexpr_compiler(CommonTokenStream(lexer))
    parser.deep = deep
    parser.message = message
    if error_reporter:
        # default options (state_only = False)
        parser.set_error_reporter(error_reporter)
    # ident checking option: only state and input names are allowed
    parser.state_only = True
    try:
        tree = parser.sig_bool(symb_table)
    except RecognitionException as exc:
        if error_reporter:
            error_reporter.display("Error in condition: %s" % exc)
            parser.displayExceptionMessage(exc)
            return None
        else:
            raise exc
    return tree.exp


def compile_constraint(const_txt, symb_table, error_reporter):
    """Compile a constraint expression to a tree representation

    @param const_txt: string to be compiled
    @param symb_table: symbol table of the compiler
    @param error_reporter: output error messages
    @return: a tree representing the expression
    """
    text_c = const_txt + " $"
    reader = ANTLRStringStream(text_c)
    lexer = sigexpr_lexer(reader)
    parser = sigexpr_compiler(CommonTokenStream(lexer))
    parser.set_error_reporter(error_reporter)
    try:
        tree = parser.sig_expression(symb_table)
    except RecognitionException as exc:
        error_reporter.display("Error in constraints: %s" % exc)
        parser.displayExceptionMessage(exc)
        return None
    return tree.exp


def compile_event(text, symb_table, free_clocks, error_reporter):
    """Compile an event expression to a tree representation

    @param text: string to be compiled
    @param symb_table: symbol table of the compiler
    @param free_clocks: boolean if true free clocks are collected
    @param error_reporter: output error messages
    @return: a triple exp, se, fc
        where exp is the event expression,
        se the state events (s#> ..) used and
        fc is the free clocks used in the event expression.
    """
    text_c = text + " $"
    reader = ANTLRStringStream(text_c)
    lexer = sigexpr_lexer(reader)
    # default options on compiler (state_only = False)
    parser = sigexpr_compiler(CommonTokenStream(lexer))
    parser.set_error_reporter(error_reporter)
    parser.catch_free_clocks = free_clocks
    # ident checking option: only state and input names are allowed in bool op
    parser.state_only_in_bool = True
    try:
        tree = parser.sig_expression(symb_table)
    except RecognitionException as exc:
        error_reporter.display("Error in event: %s" % exc)
        parser.displayExceptionMessage(exc)
        return None
    return tree.exp, parser.state_events, parser.free_clocks



# -*- coding: utf-8 -*-
## Filename    : TestCLUnfolder.py
## Author(s)   : Michel Le Borgne
## Created     : 13 sept. 2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s):
##
"""Unitary Test of the unfolder"""
from __future__ import print_function
import unittest
import pytest
import pkg_resources

from cadbiom.models.guard_transitions.chart_model import ChartModel
from cadbiom.models.clause_constraints.CLDynSys import Literal, Clause
from cadbiom.models.guard_transitions.analyser.ana_visitors import TableVisitor
from cadbiom.models.clause_constraints.CLDynSys import CLDynSys
from cadbiom.models.clause_constraints.mcl.MCLTranslators import GT2Clauses
from cadbiom.models.clause_constraints.mcl.CLUnfolder import CLUnfolder

from cadbiom.models.clause_constraints.mcl.MCLAnalyser import MCLAnalyser
from cadbiom.models.clause_constraints.mcl.MCLQuery import MCLSimpleQuery


class ErrorReporter(object):
    """error reporter of the compil type"""
    def __init__(self):
        self.context = ""
        self.error = False
        pass

    def display(self, mess):
        """
        just printing
        """
        self.error = True
        print('\n>> '+self.context+"  "+mess)

    def display_info(self, mess):
        """
        just printing
        """
        print('\n-- '+mess)

    def set_context(self, cont):
        """
        for transition compiling
        """
        self.context = cont

# helper functions

def string_to_clause(text_clause):
    """
    text_clause is a string such that "a,not b,not c"
    it is transformed in the corresponding clause
    WARNING: strict syntax - no check - for tests
    """
    # split into literals
    clause = Clause([])
    llit = text_clause.split(',')
    # translate each literal
    for lit in llit:
        spl = lit.split()
        if len(spl) == 1: # "a"
            clause.add_lit(Literal(spl[0], True))
        else: # "not a"
            clause.add_lit(Literal(spl[1], False))
    return clause

def model1():
    """
    A simple ChartModel with two nodes and a transition
    """
    # build dynamical system
    model = ChartModel("Test")
    root = model.get_root()
    node_1 = root.add_simple_node('n1', 0, 0)
    node_2 = root.add_simple_node('n2', 0, 0)
    root.add_transition(node_1, node_2)
    return model

def model2():
    """
    ChartModel model1 + 3 nodes and 2 transitions and a cycle without start
    """
    # build dynamical system
    model = model1()
    root = model.get_root()
    node_3 = root.add_simple_node('n3', 0, 0)
    node_4 = root.add_simple_node('n4', 0, 0)
    root.add_transition(node_3, node_4)
    node_5 = root.add_simple_node('n5', 0, 0)
    root.add_transition(node_4, node_5)
    root.add_transition(node_5, node_3)
    return model

def model3():
    """
    ChartModel model2 + 1 node + 1 start node and a transition
    """
    # build dynamical system
    model = model2()
    root = model.get_root()
    node_s = root.add_start_node('n3', 0, 0)
    node_3 = model.get_simple_node('n3')
    root.add_transition(node_s, node_3)
    return model

def model4():
    """
    ChartModel with:
        - 1 input node: in1
        - 5 nodes: n1, n2, n3, n4, n5
        - 5 transitions: n1-n2, n3-n4, n4-n5, n5-n3, in1-n1
        - 2 free clocks: hh1: (n3 or n4) on n1-n2, hh2: (n1 and n3) on n4-n5
    """
    model = ChartModel("Test")
    root = model.get_root()
    node_1 = root.add_simple_node('n1', 0, 0)
    node_2 = root.add_simple_node('n2', 0, 0)
    tr0 = root.add_transition(node_1, node_2)
    tr0.set_event('hh1')
    tr0.set_condition("n3 or n4")
    node_3 = root.add_simple_node('n3', 0, 0)
    node_4 = root.add_simple_node('n4', 0, 0)
    root.add_transition(node_3, node_4)
    node_5 = root.add_simple_node('n5', 0, 0)
    tr1 = root.add_transition(node_4, node_5)
    root.add_transition(node_5, node_3)
    tr1.set_event("hh2")
    tr1.set_condition("n1 and n3")
    node_i =  root.add_input_node('in1', 0, 0)
    tri = root.add_transition(node_i, node_1)
    tri.set_condition("n5")

    # Frontier test: Add a start node on n3 or do not attempt any frontier
    # place because there is a SCC composed of n3, n4, n5.
    #node_s = root.add_start_node('s1', 0, 0)
    #root.add_transition(node_s, node_3)

    return model

def create_unfolder(model):
    """Return an unfolder for the given model"""
    tvisit = TableVisitor(None) # no error display
    model.accept(tvisit)
    cld = CLDynSys( tvisit.tab_symb, None)
    reporter = ErrorReporter()
    cvisit = GT2Clauses(cld, reporter, True)
    model.accept(cvisit)
    # unfolder
    return CLUnfolder(cld)

@pytest.fixture()
def textual_properties():
    """start, invariant, final"""
    return (
        ("M", "L", "C"), # No solution (M activated)
        ("", "L", "C and K"), # Solution: I E D F
        ("", "", "C"), # Solutions: F E L D, I E D F
    )

@pytest.fixture()
def numerical_properties():

    return (
        ([[13]], [[12]], [[3]]),
        ([], [[12]], [[3, -47], [11, -47], [-3, -11, 47], [47]]),
    )

@pytest.fixture()
def feed_mclanalyser():
    """Setup MCLAnalyser with a bcx model"""
    rep = ErrorReporter()
    # Force __var_list to be sorted at the output of base_var_set from
    # CLDynSys for reproductibility in tests (literals naming)
    mcla = MCLAnalyser(rep, debug=True)
    filename = pkg_resources.resource_filename(
        __name__, # package name
        "examples/example_model.bcx"
    )
    mcla.build_from_chart_file(filename)
    return mcla


class TestCLUnfolder(unittest.TestCase):
    """
    Test public and some private methods (test_method form)
    """
    def test_var_name(self):
        """
        Test of variables at initial state uncoding and decoding
        """
        model = model1()
        unfolder = create_unfolder(model)

        # Test unfolder init and internal dynamic system
        assert unfolder._CLUnfolder__shift_step == 2
        assert unfolder.shift_step_init == 2
        assert unfolder.get_var_number() == 2

        # naming and coding variables
        cn1 = unfolder.var_dimacs_code('n1')
        cn2 = unfolder.var_dimacs_code('n2')

        res = unfolder.get_var_name(cn1) == 'n1'
        res = res and (unfolder.get_var_name(cn2) == 'n2')
        self.assert_(res,'Error in variable name 1')

        res = unfolder.get_var_name(7) == 'n1'
        self.assert_(res,'Error in variable name 2')
        res = unfolder.get_var_name(8) == 'n2'
        self.assert_(res,'Error in variable name 3')

        res = unfolder.get_var_indexed_name(7) == 'n1_3'
        self.assert_(res,'Error in variable name 4')
        res = unfolder.get_var_indexed_name(8) == 'n2_3'
        self.assert_(res,'Error in variable name 5')

        # Test number of variables
        assert unfolder.get_system_var_number() == unfolder.get_var_number()


    def test_frontier(self):
        """
        Test frontier computation and encoding
        """
        ## model1
        model = model1()
        unfolder = create_unfolder(model)
        # test frontier: should be n1
        frontier_value = unfolder.frontier_values[0]
        res = unfolder.get_var_name(frontier_value) == 'n1'
        res = res and (len(unfolder.frontier_values) == 1)
        self.assert_(res,'Error in frontier: model1')

        ## model2 (one cycle without start)
        model = model2()
        unfolder = create_unfolder(model)

        # test frontier: should be n1
        frontier_value = unfolder.frontier_values[0]
        res = unfolder.get_var_name(frontier_value) == 'n1'
        res = res and (len(unfolder.frontier_values) == 1)
        self.assert_(res,'Error in frontier: model2')

        ## model3: same as model2 but with a start on n3
        model = model3()
        unfolder = create_unfolder(model)

        # test frontier - should be {n1, n3}
        frontier_values = unfolder.frontier_values
        res = len(frontier_values) == 2
        res = res and unfolder.get_var_name(frontier_values[0]) == 'n1'
        res = res and unfolder.get_var_name(frontier_values[1]) == 'n3'
        self.assert_(res,'Error in frontier: model3')

        ## model4 - No frontiers (even if we can get input node in1 in the solutions)
        model = model4()
        unfolder = create_unfolder(model)
        res = len(unfolder.frontier_values) == 0
        self.assert_(res,'Error in frontier: model4')


    def test_free_clocks_inputs(self):
        """
        test on free clocks and input computation
        """
        # model3: no free clock, no input
        model = model3()
        unfolder = create_unfolder(model)
        lfc = unfolder.get_free_clocks()
        res = len(lfc) == 0
        self.assert_(res,'Error in free clocks: model3')
        lin = unfolder.get_inputs()
        res = len(lin) == 0
        self.assert_(res,'Error in inputs 1')

        # model4: two free clocks and one input
        model = model4()
        unfolder = create_unfolder(model)
        lfc = unfolder.get_free_clocks()
        res = len(lfc) == 2
        found_names = {unfolder.get_var_name(clock) for clock in lfc}
        res = found_names == {'hh2', 'hh1'}
        self.assert_(res,'Error in free clocks: model4')
        lin = unfolder.get_inputs()
        res = len(lin) == 1
        found_names = {unfolder.get_var_name(inpt) for inpt in lin}
        res = found_names == {'in1'}
        self.assert_(res,'Error in inputs: model4')




def init_forward_unfolding_part_1(unfolder):
    """Initialization step only"""

    unfolder._CLUnfolder__shift_direction = 'FORWARD'
    unfolder._CLUnfolder__current_step = 1
    unfolder._CLUnfolder__shift_step = unfolder.shift_step_init  # back to basic!
    unfolder._CLUnfolder__aux_code_table = dict()            # flush auxiliary variables
    unfolder._CLUnfolder__aux_list = []                      # idem

    # Init properties to generate all variable num codes
    unfolder._CLUnfolder__init_initial_constraint_0()
    unfolder._CLUnfolder__init_final_constraint_0()
    unfolder._CLUnfolder__init_invariant_constraint_0()
    unfolder._CLUnfolder__init_variant_constraints_0()

def init_forward_unfolding_part_2(unfolder):
    """Shift step only"""

    # Now shifting is possible
    unfolder._CLUnfolder__forward_init_dynamic()
    unfolder._CLUnfolder__shift_final()
    unfolder._CLUnfolder__shift_invariant()
    # PS: __variant_constraints si already initialized for the first step.

def test_init_unfolder(feed_mclanalyser):

    mcla = feed_mclanalyser
    unfolder = mcla.unfolder # shortcut

    # Initialization step only
    init_forward_unfolding_part_1(mcla.unfolder)

    ## Test CLDynSys
    # 15 places
    assert unfolder.get_var_number() == 46

    # __var_list is built on casted set; MCLAnalyser is un debug mode
    # so the list is sorted
    expected = [
        '##',
        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'P',
        '_h2', '_h3', '_h4', '_h5', '_h6', '_h7', '_h_0', '_h_1', '_lit0', '_lit1',
        '_lit10', '_lit11', '_lit12', '_lit13', '_lit14', '_lit15', '_lit16',
        '_lit17', '_lit18', '_lit19', '_lit2', '_lit20', '_lit21', '_lit22',
        '_lit3', '_lit4', '_lit5', '_lit6', '_lit7', '_lit8', '_lit9'
    ]

    print(unfolder._CLUnfolder__var_list)
    assert unfolder._CLUnfolder__var_list == expected

    ## Test mapping of variables names to values
    expected_var_code_table = {
        '_lit22': 39, '_lit20': 37, '_lit21': 38, '_lit19': 35,
        '_lit18': 34, '_lit6': 43, '_lit17': 33, '_lit13': 29, '_lit9': 46,
        '_lit8': 45, '_h_0': 22, '_h_1': 23, '_lit16': 32, '_lit12': 28,
        '_lit3': 40, '_lit2': 36, '_lit1': 25, '_lit0': 24, '_lit7': 44,
        '_lit11': 27, '_lit5': 42, '_lit4': 41,
        'A': 1, 'C': 3, 'B': 2, 'E': 5, 'D': 4, 'G': 7, 'F': 6, 'I': 9, 'H': 8,
        'K': 11, 'J': 10, 'M': 13, 'L': 12, '_lit15': 31, 'N': 14, 'P': 15,
        '_lit14': 30, '_h4': 18, '_h5': 19, '_h6': 20, '_h7': 21, '_h2': 16,
        '_h3': 17, '_lit10': 26
    }

    # Check presence of all basic literals (not future ones with "`" added)
    found_var_code_table = list(unfolder._CLUnfolder__var_code_table.items())
    print(found_var_code_table)
    for lit_name_val in expected_var_code_table.items():
        assert lit_name_val in found_var_code_table

    # var_code_table is twice as big as __var_list items
    # (future literals with "`" are added)
    assert len(found_var_code_table) == 2 * len(expected_var_code_table)

    ## Test frontiers, places, values and names
    print("Frontier values", unfolder.frontier_values)
    assert set(unfolder.frontier_values) == {4, 5, 6, 7, 9, 12, 14} # D E F G I L N

    frontiers_values_mapping = {
        value: unfolder.get_var_name(value) for value in unfolder.frontier_values
    }
    print("Frontiers values mapping", frontiers_values_mapping)
    assert frontiers_values_mapping == {4: 'D', 5: 'E', 6: 'F', 7: 'G', 9: 'I', 12: 'L', 14: 'N'}

    found = unfolder.frontiers_negative_values
    print("Frontiers negatives values", found)
    assert found == frozenset([-14, -12, -9, -7, -6, -5, -4])

    found = unfolder.frontiers_pos_and_neg
    assert found == frozenset([4, 5, 6, 7, 9, 12, 14, -14, -12, -9, -7, -6, -5, -4])

    # Check clocks
    free_clocks = unfolder._CLUnfolder__free_clocks
    print("Free clocks/events", free_clocks)
    assert free_clocks == frozenset([16, 17, 18, 19, 20, 21, 22, 23])

    clocks_values_mapping = {
        value: unfolder.get_var_name(value) for value in unfolder._CLUnfolder__free_clocks
    }
    print("Clocks values mapping", clocks_values_mapping)
    assert clocks_values_mapping == {16: '_h2', 17: '_h3', 18: '_h4', 19: '_h5', 20: '_h6', 21: '_h7', 22: '_h_0', 23: '_h_1'}

    # Check input places
    assert unfolder._CLUnfolder__inputs == frozenset()

    # Not frontiers: A B C H J K M P
    found = unfolder._CLUnfolder__no_frontier_init
    print("Places not frontiers", found)
    assert found == [[-1], [-2], [-3], [-8], [-10], [-11], [-13], [-15]]



def test_init_forward(feed_mclanalyser, textual_properties, numerical_properties):
    """Test forward initialization for various models and properties types

    Specific call:
    pytest cadbiom/models/clause_constraints/mcl/TestCLUnfolder.py::test_init_forward

    Query attributes:
    self.dim_start = []         # list<DClause>
    self.dim_inv = []           # list<DClause>
    self.dim_final = []         # list<DClause>
    self.dim_variant_prop = []  # list<list<DClause>>

    CLUnfolder attributes that contain query attributes:
    self.__initial_property = None            # logical formula - literal boolean expression
    self.__dimacs_initial = None              # list of DIMACS clauses
    self.__final_property = None              # logical formula
    self.__dimacs_final = None                # list of DIMACS clauses
    self.__invariant_property = None          # logical formula
    self.__dimacs_invariant = None            # list of DIMACS clauses
    self.__variant_property = None            # list<logic formulas>
    self.__dimacs_variant = None              # list<list<DIMACS clauses>>

    CLUnfolder attributes that contain clauses:
    self.__list_variant_constraints = None    # list<list<DIMACS clauses>>

    dynamic_constraints => __dynamic_constraints
    initial_constraints => __initial_constraints
    invariant_constraints => __invariant_constraints
    variant_constraints => __variant_constraints
    final_constraints => __final_constraints

    Mangling prefix: unfolder._CLUnfolder

    TODO:
        - search solutions for the given properties
        - test variant properties
    """
    mcla = feed_mclanalyser

    ## First properties
    ### Textual properties
    start, invariant, final = textual_properties[0]

    query = MCLSimpleQuery(start, invariant, final)
    mcla.unfolder.init_with_query(query)

    init_forward_unfolding_solution_1(mcla)
    ## TODO: search sol

    ### DIMACS properties
    dim_start, dim_inv, dim_final = numerical_properties[0]
    query = MCLSimpleQuery(None, None, None)
    query.dim_start = dim_start
    query.dim_inv = dim_inv
    query.dim_final = dim_final
    mcla.unfolder.init_with_query(query)
    init_forward_unfolding_solution_1(mcla)


    ## Second properties
    ### Textual properties
    start, invariant, final = textual_properties[1]

    query = MCLSimpleQuery(start, invariant, final)
    mcla.unfolder.init_with_query(query)

    init_forward_unfolding_solution_2(mcla)
    ## TODO: search sol

    ### DIMACS properties: TODO: marche pas: les litéraux supplémentaires ne sont pas pris en compte...
    dim_start, dim_inv, dim_final = numerical_properties[1]
    query = MCLSimpleQuery(None, None, None)
    query.dim_start = dim_start
    query.dim_inv = dim_inv
    query.dim_final = dim_final
    mcla.unfolder.init_with_query(query)
    init_forward_unfolding_solution_2(mcla)



def init_forward_unfolding_solution_1(mcla):
    """
    - Test first part of init_forward_unfolding: init of constraints
    - Test second part of init_forward_unfolding: shift of initialized constraints

    ("M", "L", "C"), # No solution (M activated)
    """
    unfolder = mcla.unfolder # shortcut
    init_forward_unfolding_part_1(mcla.unfolder)

    ############################################################################

    print(unfolder.dynamic_constraints)
    print(unfolder.initial_constraints)
    print(unfolder.invariant_constraints)
    print(unfolder.variant_constraints)
    print(unfolder.final_constraints)

    # No auxiliary variables
    assert unfolder._CLUnfolder__aux_code_table == dict()
    assert unfolder._CLUnfolder__aux_list == []

    # MCLA could be reused, so __dynamic_constraints is not empty at this step
    # artificial reset
    # unfolder._CLUnfolder__dynamic_constraints = []
    # assert unfolder.dynamic_constraints == []

    # no frontiers + M (start place)
    assert unfolder.initial_constraints == [[-1], [-2], [-3], [-8], [-10], [-11], [-13], [-15], [13]]

    # L
    assert unfolder.invariant_constraints == [[[12]]]

    # No variant constraint
    assert unfolder.variant_constraints == []

    # C
    assert unfolder.final_constraints == [[3]]

    ############################################################################

    init_forward_unfolding_part_2(unfolder)

    print(unfolder.dynamic_constraints)
    print(unfolder.invariant_constraints)
    print(unfolder.final_constraints)

    # Shift of the system clauses (and syst aux clauses)
    # The maximum value is 46 since there is 46 literals
    expected = [
        [[-24, 16], [-24, 4], [-16, -4, 24], [-15, 42], [-12, 42], [15, 12, -42],
        [-42, 41], [-11, 41], [42, 11, -41], [2, -40], [41, -40], [-2, -41, 40],
        [40, -36], [-13, -36], [-40, 13, 36], [1, -25], [36, -25], [-1, -36, 25],
        [-43, 22], [-43, 25], [-22, -25, 43], [-47, 24, 1], [-47, 24, -43],
        [47, -24], [47, -1, 43], [5, -44], [6, -44], [-5, -6, 44], [-45, 17],
        [-45, 44], [-17, -44, 45], [1, -27], [41, -27], [-1, -41, 27], [27, -26],
        [-13, -26], [-27, 13, 26], [2, -46], [26, -46], [-2, -26, 46], [-30, 22],
        [-30, 46], [-22, -46, 30], [-48, 45, 2], [-48, 45, -30], [48, -45],
        [48, -2, 30], [31, -30], [31, -43], [30, 43, -31], [-49, 31, 3], [49, -31],
        [49, -3], [-50, 4], [-50, -24], [50, -4, 24], [-51, 5], [-51, -45],
        [51, -5, 45], [-52, 6], [52, -6], [-32, 21], [-32, 7], [-21, -7, 32],
        [-53, 7], [-53, -32], [53, -7, 32], [-54, 32, 8], [54, -32], [54, -8],
        [-33, 20], [-33, 11], [-20, -11, 33], [-34, 18], [-34, 9], [-18, -9, 34],
        [-55, 33, 9], [-55, 33, -34], [55, -33], [55, -9, 34], [-35, 19], [-35, 10],
        [-19, -10, 35], [-56, 34, 10], [-56, 34, -35], [56, -34], [56, -10, 35],
        [-57, 35, 11], [-57, 35, -33], [57, -35], [57, -11, 33], [-58, 12],
        [58, -12], [-37, 23], [-37, 14], [-23, -14, 37], [-38, 23], [-38, 14],
        [-23, -14, 38], [39, -37], [39, -38], [37, 38, -39], [-60, 14], [-60, -39],
        [60, -14, 39], [-59, 37, 13], [59, -37], [59, -13], [-61, 38, 15], [61, -38],
        [61, -15], [-18, 9], [-19, 10], [-20, 11], [-21, 7], [-16, 4], [-17, 5],
        [-22, 1, 2], [-23, 14, 14]]]

    # Crap
    assert unfolder.dynamic_constraints == expected

    # L
    assert unfolder.invariant_constraints == [[[12]], [[58]]]

    # No variant constraint (not shifted in second part of init_forward_unfolding)
    assert unfolder.variant_constraints == []

    # C + 46
    assert unfolder.final_constraints == [[49]]


def init_forward_unfolding_solution_2(mcla):
    """
    - Test first part of init_forward_unfolding: init of constraints
    - Test second part of init_forward_unfolding: shift of initialized constraints

    ("", "L", "C and K") # I E D F
    """
    unfolder = mcla.unfolder # shortcut
    init_forward_unfolding_part_1(mcla.unfolder)

    ############################################################################

    print(unfolder.dynamic_constraints)
    print(unfolder.initial_constraints)
    print(unfolder.invariant_constraints)
    print(unfolder.variant_constraints)
    print(unfolder.final_constraints)

    # No auxiliary variables: "C and K" property is added
    assert unfolder._CLUnfolder__aux_code_table == {'_lit47': 47}
    assert unfolder._CLUnfolder__aux_list == ['_lit47']

    # MCLA could be reused, so __dynamic_constraints is not empty at this step
    # artificial reset
    # unfolder._CLUnfolder__dynamic_constraints = []
    # assert unfolder.dynamic_constraints == []

    # no frontiers + nothing (no start place)
    assert unfolder.initial_constraints == [[-1], [-2], [-3], [-8], [-10], [-11], [-13], [-15]]

    # L
    assert unfolder.invariant_constraints == [[[12]]]

    # No variant constraint
    assert unfolder.variant_constraints == []

    # "C and K" = _lit47
    # C, not _lit47
    # K, not _lit47
    # not C, not K, _lit47
    # _lit47
    assert unfolder.final_constraints == [[3, -47], [11, -47], [-3, -11, 47], [47]]

    # PS:
    # "C or K" = _lit47
    # not C, _lit47
    # not K, _lit47
    # C, K, not _lit47
    # _lit47
    #[[-3, 47], [-11, 47], [3, 11, -47], [47]]

    ############################################################################

    init_forward_unfolding_part_2(unfolder)

    print(unfolder.dynamic_constraints)
    print(unfolder.invariant_constraints)
    print(unfolder.final_constraints)

    # Shift of the system clauses (and syst aux clauses)
    # The maximum value is now 48 since there is 47 literals
    assert unfolder.get_var_number() == 46
    assert unfolder.get_shift_step() == 47
    expected = [
        [[-24, 16], [-24, 4], [-16, -4, 24], [-15, 42], [-12, 42], [15, 12, -42],
        [-42, 41], [-11, 41], [42, 11, -41], [2, -40], [41, -40], [-2, -41, 40],
        [40, -36], [-13, -36], [-40, 13, 36], [1, -25], [36, -25], [-1, -36, 25],
        [-43, 22], [-43, 25], [-22, -25, 43], [-48, 24, 1], [-48, 24, -43],
        [48, -24], [48, -1, 43], [5, -44], [6, -44], [-5, -6, 44], [-45, 17],
        [-45, 44], [-17, -44, 45], [1, -27], [41, -27], [-1, -41, 27], [27, -26],
        [-13, -26], [-27, 13, 26], [2, -46], [26, -46], [-2, -26, 46], [-30, 22],
        [-30, 46], [-22, -46, 30], [-49, 45, 2], [-49, 45, -30], [49, -45],
        [49, -2, 30], [31, -30], [31, -43], [30, 43, -31], [-50, 31, 3], [50, -31],
        [50, -3], [-51, 4], [-51, -24], [51, -4, 24], [-52, 5], [-52, -45],
        [52, -5, 45], [-53, 6], [53, -6], [-32, 21], [-32, 7], [-21, -7, 32],
        [-54, 7], [-54, -32], [54, -7, 32], [-55, 32, 8], [55, -32], [55, -8],
        [-33, 20], [-33, 11], [-20, -11, 33], [-34, 18], [-34, 9], [-18, -9, 34],
        [-56, 33, 9], [-56, 33, -34], [56, -33], [56, -9, 34], [-35, 19], [-35, 10],
        [-19, -10, 35], [-57, 34, 10], [-57, 34, -35], [57, -34], [57, -10, 35],
        [-58, 35, 11], [-58, 35, -33], [58, -35], [58, -11, 33], [-59, 12],
        [59, -12], [-37, 23], [-37, 14], [-23, -14, 37], [-38, 23], [-38, 14],
        [-23, -14, 38], [39, -37], [39, -38], [37, 38, -39], [-61, 14], [-61, -39],
        [61, -14, 39], [-60, 37, 13], [60, -37], [60, -13], [-62, 38, 15],
        [62, -38], [62, -15], [-18, 9], [-19, 10], [-20, 11], [-21, 7], [-16, 4],
        [-17, 5], [-22, 1, 2], [-23, 14, 14]]]

    # Crap
    assert unfolder.dynamic_constraints == expected

    # L : 12 + shift_step (47)
    assert unfolder.invariant_constraints == [[[12]], [[59]]]

    # No variant constraint (not shifted in second part of init_forward_unfolding)
    assert unfolder.variant_constraints == []

    # "C and K" = _lit47 = 47
    # Shift: [[3, -47], [11, -47], [-3, -11, 47], [47]] with shift_step (47)
    assert unfolder.final_constraints == [[50, -94], [58, -94], [-50, -58, 94], [94]]


def test_shift(feed_mclanalyser, textual_properties):
    """Test shift of constraints during the solutions search"""

    mcla = feed_mclanalyser
    unfolder = mcla.unfolder # shortcut

    start, invariant, final = textual_properties[0]

    query = MCLSimpleQuery(start, invariant, final)
    unfolder.init_with_query(query)

    init_forward_unfolding_part_1(mcla.unfolder)
    init_forward_unfolding_part_2(unfolder)

    unfolder.shift()


    print(unfolder.dynamic_constraints)
    print(unfolder.initial_constraints) # Not changed
    print(unfolder.invariant_constraints)
    print(unfolder.variant_constraints) # No data for this test
    print(unfolder.final_constraints)

    expected = [
        # Initial step
        [[-24, 16], [-24, 4], [-16, -4, 24], [-15, 42], [-12, 42], [15, 12, -42],
        [-42, 41], [-11, 41], [42, 11, -41], [2, -40], [41, -40], [-2, -41, 40],
        [40, -36], [-13, -36], [-40, 13, 36], [1, -25], [36, -25], [-1, -36, 25],
        [-43, 22], [-43, 25], [-22, -25, 43], [-47, 24, 1], [-47, 24, -43],
        [47, -24], [47, -1, 43], [5, -44], [6, -44], [-5, -6, 44], [-45, 17],
        [-45, 44], [-17, -44, 45], [1, -27], [41, -27], [-1, -41, 27], [27, -26],
        [-13, -26], [-27, 13, 26], [2, -46], [26, -46], [-2, -26, 46], [-30, 22],
        [-30, 46], [-22, -46, 30], [-48, 45, 2], [-48, 45, -30], [48, -45],
        [48, -2, 30], [31, -30], [31, -43], [30, 43, -31], [-49, 31, 3], [49, -31],
        [49, -3], [-50, 4], [-50, -24], [50, -4, 24], [-51, 5], [-51, -45],
        [51, -5, 45], [-52, 6], [52, -6], [-32, 21], [-32, 7], [-21, -7, 32],
        [-53, 7], [-53, -32], [53, -7, 32], [-54, 32, 8], [54, -32], [54, -8],
        [-33, 20], [-33, 11], [-20, -11, 33], [-34, 18], [-34, 9], [-18, -9, 34],
        [-55, 33, 9], [-55, 33, -34], [55, -33], [55, -9, 34], [-35, 19], [-35, 10],
        [-19, -10, 35], [-56, 34, 10], [-56, 34, -35], [56, -34], [56, -10, 35],
        [-57, 35, 11], [-57, 35, -33], [57, -35], [57, -11, 33], [-58, 12],
        [58, -12], [-37, 23], [-37, 14], [-23, -14, 37], [-38, 23], [-38, 14],
        [-23, -14, 38], [39, -37], [39, -38], [37, 38, -39], [-60, 14], [-60, -39],
        [60, -14, 39], [-59, 37, 13], [59, -37], [59, -13], [-61, 38, 15], [61, -38],
        [61, -15], [-18, 9], [-19, 10], [-20, 11], [-21, 7], [-16, 4], [-17, 5],
        [-22, 1, 2], [-23, 14, 14]],
        # Shifted step
        [[-70, 62],
        [-70, 50], [-62, -50, 70], [-61, 88], [-58, 88], [61, 58, -88], [-88, 87],
        [-57, 87], [88, 57, -87], [48, -86], [87, -86], [-48, -87, 86], [86, -82],
        [-59, -82], [-86, 59, 82], [47, -71], [82, -71], [-47, -82, 71], [-89, 68],
        [-89, 71], [-68, -71, 89], [-93, 70, 47], [-93, 70, -89], [93, -70],
        [93, -47, 89], [51, -90], [52, -90], [-51, -52, 90], [-91, 63], [-91, 90],
        [-63, -90, 91], [47, -73], [87, -73], [-47, -87, 73], [73, -72],
        [-59, -72], [-73, 59, 72], [48, -92], [72, -92], [-48, -72, 92], [-76, 68],
        [-76, 92], [-68, -92, 76], [-94, 91, 48], [-94, 91, -76], [94, -91],
        [94, -48, 76], [77, -76], [77, -89], [76, 89, -77], [-95, 77, 49], [95, -77],
        [95, -49], [-96, 50], [-96, -70], [96, -50, 70], [-97, 51], [-97, -91],
        [97, -51, 91], [-98, 52], [98, -52], [-78, 67], [-78, 53], [-67, -53, 78],
        [-99, 53], [-99, -78], [99, -53, 78], [-100, 78, 54], [100, -78], [100, -54],
        [-79, 66], [-79, 57], [-66, -57, 79], [-80, 64], [-80, 55], [-64, -55, 80],
        [-101, 79, 55], [-101, 79, -80], [101, -79], [101, -55, 80], [-81, 65],
        [-81, 56], [-65, -56, 81], [-102, 80, 56], [-102, 80, -81], [102, -80],
        [102, -56, 81], [-103, 81, 57], [-103, 81, -79], [103, -81], [103, -57, 79],
        [-104, 58], [104, -58], [-83, 69], [-83, 60], [-69, -60, 83], [-84, 69],
        [-84, 60], [-69, -60, 84], [85, -83], [85, -84], [83, 84, -85], [-106, 60],
        [-106, -85], [106, -60, 85], [-105, 83, 59], [105, -83], [105, -59],
        [-107, 84, 61], [107, -84], [107, -61], [-64, 55], [-65, 56], [-66, 57],
        [-67, 53], [-62, 50], [-63, 51], [-68, 47, 48], [-69, 60, 60]]]

    assert unfolder.dynamic_constraints == expected

    # L = 12: + 46 + 46 => 104
    assert unfolder.invariant_constraints == [[[12]], [[58]], [[104]]]

    # C = 3: + 46 + 46 =>
    assert unfolder.final_constraints == [[95]]


if __name__ == "__main__":
    unittest.main()





# -*- coding: utf-8 -*-
"""Unit tests for Python API"""

from __future__ import unicode_literals
from __future__ import print_function


# Standard imports
import pytest
import tempfile

@pytest.fixture()
def feed_model_with_SCC():
    """Fixture for a Cadbiom model with 4 Strongly Connected Components

    :return: A list of SCC, and the corrected model with start nodes inserted.
    :rtype: <tuple <list>, <str>>
    """

    scc = [['I', 'K', 'J', 'L'], ['Y', 'X', 'Z']]
    model = """<model xmlns="http://cadbiom" name="">
  <CSimpleNode name="A" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="B" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="E" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="D" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="C" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="Y" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="X" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="Z" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="I" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="K" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="J" xloc="0.0" yloc="0.0"/>
  <CSimpleNode name="L" xloc="0.0" yloc="0.0"/>
  <CStartNode name="__start__0" xloc="0" yloc="0"/>
  <CStartNode name="__start__1" xloc="0" yloc="0"/>
  <transition ori="I" ext="J" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="L" ext="I" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="K" ext="L" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="X" ext="Z" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="__start__1" ext="X" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="J" ext="K" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="B" ext="D" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="A" ext="B" event="" condition="C" action="" fact_ids="[]"/>
  <transition ori="X" ext="C" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="C" ext="B" event="" condition="A" action="" fact_ids="[]"/>
  <transition ori="I" ext="K" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="D" ext="C" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="J" ext="B" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="__start__0" ext="I" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="Y" ext="X" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="Z" ext="Y" event="" condition="" action="" fact_ids="[]"/>
  <transition ori="D" ext="E" event="" condition="" action="" fact_ids="[]"/>
</model>
"""

    return scc, model


def test_SCC_search(feed_model_with_SCC):
    """
    """

    import cadbiom.models.guard_transitions.analyser.model_corrections as mc

    model = feed_model_with_SCC[1]

    # Create the file model in /tmp/
    # Note: prevent the deletion of the file after the close() call
    fd_model = tempfile.NamedTemporaryFile(suffix='.bcx', delete=False)
    fd_model.write(
        """<model xmlns="http://cadbiom" name="">
    <CSimpleNode name="A"/>
    <CSimpleNode name="B"/>
    <CSimpleNode name="E"/>
    <CSimpleNode name="D"/>
    <CSimpleNode name="C"/>
    <CSimpleNode name="Y"/>
    <CSimpleNode name="X"/>
    <CSimpleNode name="Z"/>
    <CSimpleNode name="I"/>
    <CSimpleNode name="K"/>
    <CSimpleNode name="J"/>
    <CSimpleNode name="L"/>
    <transition name="" ori="A" ext="B" event="" condition="C"/>
    <transition name="" ori="B" ext="D" event="" condition=""/>
    <transition name="" ori="C" ext="B" event="" condition="A"/>
    <transition name="" ori="D" ext="C" event="" condition=""/>
    <transition name="" ori="D" ext="E" event="" condition=""/>
    <transition name="" ori="Z" ext="Y" event="" condition=""/>
    <transition name="" ori="Y" ext="X" event="" condition=""/>
    <transition name="" ori="X" ext="Z" event="" condition=""/>
    <transition name="" ori="X" ext="C" event="" condition=""/>
    <transition name="" ori="I" ext="J" event="" condition=""/>
    <transition name="" ori="L" ext="I" event="" condition=""/>
    <transition name="" ori="K" ext="L" event="" condition=""/>
    <transition name="" ori="J" ext="K" event="" condition=""/>
    <transition name="" ori="J" ext="B" event="" condition=""/>
    <transition name="" ori="I" ext="K" event="" condition=""/>
    </model>"""
    )
    fd_model.close()


    # Make a new model file (with "_without_scc" suffix in filename)
    mc.add_start_nodes(fd_model.name) # Filename + path

    with open(fd_model.name[:-4] + "_without_scc.bcx", 'r') as file:
        assert model == file.read()



## Filename    : simul_exceptions.py
## Author(s)   : Michel Le Borgne
## Created     : 04/2010
## Revision    : 
## Source      : 
##
## Copyright 2010 : IRISA/INRIA
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and INRIA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall INRIA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## INRIA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA/INRIA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE 
##     
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##
"""
Simulator exceptions
"""

class SimulException(Exception):
    """
    Exceptions for simulation management
    """
    def __init__(self, mess):
        self.message = mess
        
    def __str__(self):
        return self.message

class SimulStopException(SimulException):
    """
    End of a simulation (mostly end of input)
    """
    def __init__(self):
        self.message = 'STOP'
    
    def __str__(self):
        return self.message
    
class SimulPropViolation(SimulException):
    """
    Used when an invariant property or a variant one if violated
    """
    def __init__(self, mess, prop):
        self.message = mess + " : "+prop
    
    def __str__(self):
        return self.message




# -*- coding: utf-8 -*-
## Filename    : chart_xml.py
## Author(s)   : Geoffroy Andrieux
## Created     : 04/2010
## Revision    :
## Source      :
##
## Copyright 2010 : IRISA
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA/IRSET
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Michel Le Borgne, Nolwenn Le Meur
##
"""
Load and generate Cadbiom xml files
"""
from __future__ import unicode_literals
from __future__ import print_function

from cadbiom.models.guard_transitions.chart_model import ChartModel
from xml.sax import make_parser
from xml.sax import parseString as PS
from xml.sax.handler import ContentHandler
from lxml import etree
from lxml import objectify


class XmlVisitor:
    """Visitor used to generate xml cadbiom code when the model is exported."""

    def __init__(self, model):
        self.model = model
        self.fact_list = []
        self.xml = ""      # string: xml representation of model
        self.symb = dict() # symbol table to check double naming of nodes
        self.visit_chart_model()

    def visit_chart_model(self):
        """
        Entrance point
        """
        self.visit_ctop_node(self.model.get_root())

    def check_name(self, name):
        """
        Detect double declarations
        """
        try:
            self.symb[name]
        except:
            self.symb[name] = "ok"
            return
        raise XmlException("Node double declaration")

    def visit_cstart_node(self, snode):
        """
        Generate xml representation of a start node
        """

        tag = "CStartNode"
        attrname = ["name", "xloc", "yloc"]

        attr = [snode.name, snode.xloc, snode.yloc]
        return [tag, attrname, attr]

    def visit_ctrap_node(self, tnode):
        """
        Generate xml representation of a trap node
        """
        tag = "CTrapNode"
        attrname = ["name", "xloc", "yloc"]
        attr = [tnode.name, tnode.xloc, tnode.yloc]
        return [tag, attrname, attr]

    def visit_csimple_node(self, sin):
        """
        Generate xml representation of a simple node
        """
        self.check_name(sin.name)
        tag = "CSimpleNode"
        attrname = ["name", "xloc", "yloc"]
        attr = [sin.name, sin.xloc, sin.yloc]
        return [tag, attrname, attr]

    def visit_cperm_node(self, pnode):
        """
        Generate xml representation of a perm node
        """
        self.check_name(pnode.name)
        tag = "CPermNode"
        attrname = ["name", "xloc", "yloc"]
        attr = [pnode.name, pnode.xloc, pnode.yloc]
        return [tag, attrname, attr]

    def visit_cinput_node(self, inn):
        """
        Generate xml representation of an input node
        """
        # double declaration of input nodes is allowed"
        tag = "CInputNode"
        attrname = ["name", "xloc", "yloc"]
        attr = [inn.name, inn.xloc, inn.yloc]
        return [tag, attrname, attr]


    def visit_cmacro_node(self, mnode):
        """
        Generate xml representation of a macro node
        """
        self.check_name(mnode.name)
        save_macro = self.current_element
        tag = "CMacroNode"
        attrname = ["name", "xloc", "yloc", "wloc", "hloc"]
        attr = [mnode.name, mnode.xloc, mnode.yloc, mnode.wloc, mnode.hloc]
        properties = [tag, attrname, attr]

        macro = etree.SubElement(self.current_element, properties[0])
        self.current_element = macro
        if len(properties) > 1:
            attrname = properties[1]
            attr = properties[2]
            attributes = macro.attrib
            for i in range(0, len(attrname)):
                attributes[attrname[i]] =  str(attr[i])

        # nodes
        for snode in mnode.sub_nodes:
            properties = snode.accept(self)
            if properties[0] == 'CMacroNode':
                self.current_element = macro

            if properties[0] != 'CMacroNode':
                subel = etree.SubElement(self.current_element, properties[0])
                if len(properties) > 1:
                    attrname = properties[1]
                    attr = properties[2]
                    attributes = subel.attrib
                    for i in range(0, len(attrname)):
                        attributes[attrname[i]] =  str(attr[i])

        # transitions
        for gtr in mnode.transitions:
            for trunlist in gtr:
                properties = trunlist.accept(self)
                sub_tr = etree.SubElement(self.current_element, properties[0])
                if len(properties) > 1:
                    attrname = properties[1]
                    attr = properties[2]
                    attributes = sub_tr.attrib
                    for i in range(0, len(attrname)):
                        attributes[attrname[i]] =  str(attr[i])
        self.current_element = save_macro
        return [tag, attrname, attr]

    def visit_ctop_node(self, tnode):
        """Interative build of xml tree for model saving

        .. note:: namespace seems to be useless regarding nsmap here,
        because we use the default namespace without prefix...
        See http://lxml.de/tutorial.html#namespaces.
        """
        header = objectify.ElementMaker(
            annotate=False,
            #namespace="http://cadbiom.genouest.org/",
            #nsmap={None: "http://cadbiom.genouest.org/"}
            namespace=self.model.xml_namespace,
            # the default namespace (no prefix)
            nsmap={None: self.model.xml_namespace}
        )
        xmodel = header.model(name=self.model.name)
        self.current_element = xmodel

        def create_xml_element(entity):
            """Create XML element and add it to root object"""
            # get node or transition properties
            properties = entity.accept(self)
            if properties[0] != 'CMacroNode':
                element = etree.Element(properties[0])
                if len(properties) > 1:
                    attrname = properties[1]
                    attr = properties[2]
                    attributes = element.attrib
                    # Set attributes and values (name, event, coords...)
                    for i in range(0, len(attrname)):
                        attributes[attrname[i]] = str(attr[i])
                # Add notes/text of element
                if entity.note:
                    element.text = entity.note
                # Attach element to the model
                xmodel.append(element)

        # nodes
        for snode in tnode.sub_nodes:
            create_xml_element(snode)

        # transitions
        for gtr in tnode.transitions:
            for trans in gtr:
                create_xml_element(trans)

        # constraints
        if len(tnode.model.constraints) > 0:
            const = etree.Element("constraints")
            const.text = tnode.model.constraints
            xmodel.append(const)

        self.xml = etree.tostring(xmodel, pretty_print=True)
#        print (etree.tostring(xmodel,pretty_print=True))


    def visit_ctransition(self, trans):
        """
        Generate xml representation of a transition
        """
        tag = "transition"
        attrname = ["ori", "ext", "event",
                    "condition", "action", "fact_ids"]
        attr = [trans.ori.name, trans.ext.name, trans.event,
                trans.condition, trans.action, trans.fact_ids]

        fact_ids = trans.fact_ids
        for fact in fact_ids:
            self.fact_list.append(fact)

        return [tag, attrname, attr]

    def return_xml(self):
        """Return the model as xml string.

        .. note:: Used when the model is saved in a .bcx file.
        """
        return self.xml

    def get_fact_ids(self):
        """
        get litterature references
        """
        model_fact = []
        for i in self.fact_list:
            if i in model_fact:
                continue
            else :
                model_fact.append(i)
        return model_fact


class MakeHandler(ContentHandler):
    """Make a handler for the parser when the model is loaded.

    Users are expected to subclass ContentHandler to support their application.
    The following methods are called by the parser on the appropriate events
    in the input document:

    https://docs.python.org/2/library/xml.sax.handler.html
    """

    def __init__(self, model = None):
        self.pile_node = []
        self.top_pile = None
        self.pile_dict = []
        self.node_dict = dict()
        self.in_constraints = False
        self.constraints = ""
        self.model = model
        self.nodes_types = (
            'CStartNode', 'CTrapNode', 'CSimpleNode', 'CPermNode', 'CInputNode'
        )
        # Memorize the current node/transition because of inner text (note)
        # processing
        self.current_element = None

    def init_node_functions(self):
        """Bind functions to add different types of nodes to the cadbiom model

        .. note:: Must be call after the init of self.top_pile with the xml root
            object.
        """
        self.add_node_functions = \
            {
                'CStartNode': self.top_pile.add_start_node,
                'CTrapNode': self.top_pile.add_trap_node,
                'CSimpleNode': self.top_pile.add_simple_node,
                'CPermNode': self.top_pile.add_perm_node,
                'CInputNode': self.top_pile.add_input_node,
            }

    def startElement(self, name, att):
        """Signal the start of an element

        .. notes:: Nodes have to be at the top of the model (Before transitions)
            Transitions do not allow reflexive ones
            (as it could be said in the doc);
            Duplication of transitions are not authorized but only print a
            warning in the shell (they are not taken into account)

        :param arg1: Contains the raw XML 1.0 name of the element.
        :param arg2: Holds an object of the Attributes interface.
        :type arg1: <str>
        :type arg2: <xml.sax.xmlreader.AttributesImpl>
        """
        #print(att.getNames())

        if name in self.nodes_types:
            # TODO: Uniformization of API in CMacroNode() class;
            # the attribute 'name' should be at the same last position...
            element_name = att.get('name', '').encode('ascii')
            self.current_element = self.add_node_functions[name](
                name=element_name,
                xcoord=float(att.get('xloc', '0')),
                ycoord=float(att.get('yloc', '0')),
            )
            self.node_dict[element_name] = self.current_element

        elif name == 'transition':
            #name = att.get('name', '').encode('ascii')
            ori = att.get('ori', '')
            ext = att.get('ext', '')
            event = att.get('event', '')
            condition = att.get('condition', '')
            action = att.get('action', '')
            fact_ids_text = att.get('fact_ids','')[1:-1]
            if len(fact_ids_text) > 0:
                fact_ids = [int(id) for id in fact_ids_text.split(',')]
            else:
                fact_ids = []

            # Get nodes involved in the transition
            # If not present, raise an exception
            # => nodes have to be at the top of the model
            try:
                node_ori = self.node_dict[ori]
                node_ext = self.node_dict[ext]
            except Exception as exc:
                print('Bad xml file - missing nodes', ori, ' -> ', ext)
                print(self.node_dict)
                print(exc)

            self.current_element = self.top_pile.add_transition(node_ori, node_ext)
            # The transition may not be created (origin = ext for example)
            # /!\ Transitions do not allow reflexive ones
            # (as it could be said in the doc)
            # Duplication of transitions are not authorized but only print a
            # warning in the shell (they are not taken into account)
            if self.current_element:
                self.current_element.set_event(event)
                self.current_element.set_condition(condition)
                self.current_element.set_action(action)
                self.current_element.fact_ids = fact_ids

        elif name == 'CMacroNode':
            name = att.get('name', '').encode('ascii')
            xloc = float(att.get('xloc', '0'))
            yloc = float(att.get('yloc', '0'))
            wloc = float(att.get('wloc', '5'))
            hloc = float(att.get('hloc', '5'))

            node = self.top_pile.add_macro_subnode(name, xloc, yloc,
                                                   wloc, hloc)
            self.node_dict[name] = node

            self.pile_node.append(node)
            # symbol table put on stack to preserve macro scope for inputs
            new_node_dict = dict()
            self.pile_dict.append(new_node_dict)
            self.top_pile = node
            self.node_dict = new_node_dict

        elif name == 'constraints':
            self.in_constraints = True
            self.constraints = ""

        elif name == 'model':
            if not self.model:
                # Init CharModel: get name and namespace (default v1)
                self.model = ChartModel(
                    att.get('name', ''),
                    att.get('xmlns', 'http://cadbiom.genouest.org/')
                )
            # Root is a virtual macronode on top of the hierarchy.
            # A model can be a list of hierarchy grouped under this node.
            root = self.model.get_root()
            self.pile_node.append(root)
            self.top_pile = root
            self.init_node_functions()
            new_dict = dict()
            self.pile_dict.append(new_dict)
            self.node_dict = new_dict

    def characters(self, chr):
        """Receive notification of character data.

        The Parser will call this method to report each chunk of character data.
        SAX parsers may return all contiguous character data in a single chunk,
        or they may split it into several chunks;
        => we need to do a concatenation

        :param arg1: chunck of characters.
        :type arg1: <str>
        """
        # The current elem is a constraint, a transition or a node
        # print("all", self.current_element, '<'+chr+'>')
        if self.in_constraints:
            self.constraints += chr
        elif self.current_element:
            # node or transition is currently opened in startElement()
            self.current_element.note += chr

    def endElement (self, name):
        """Called when an elements ends

        .. note:: We handle only elements that need post processing like
            transitions and nodes: reset self.current_element that is used
            to load notes (inner text of xml object).
        """

        if name == 'transition' or name in self.nodes_types:
            # Close the current node or transition opened in startElement()
            self.current_element = None
        elif name == 'CMacroNode':
            #self.top_pile = self.pile_node.pop()
            self.pile_node.remove(self.top_pile)
            self.top_pile = self.pile_node[-1]
            #self.node_dict = self.pile_dict.pop()
            self.pile_dict.remove(self.node_dict)
            self.node_dict = self.pile_dict[-1]
        elif name == 'constraints':
            self.in_constraints = False
            self.model.constraints = self.constraints + '\n'
        #elif name == 'model':
        #    print(len([e for e in self.top_pile.transitions]))
        #    print(len(self.top_pile.new_transitions))


class MakeModelFromXmlFile:
    """
    parse a xml file
    """
    def __init__(self, xml_file, model = None):
        self.model = model
        self.handler = MakeHandler(model=self.model)
        self.parser = make_parser()
        self.parser.setContentHandler(self.handler)
        try:
            self.parser.parse(xml_file)
        except Exception:
            print('ERROR while xml parsing')
            raise


    def get_model(self):
        """
        As it says
        """
        return self.handler.model

class MakeModelFromXmlString:
    """
    parse a xml description as string
    """
    def __init__(self, xml_string):
        self.model = None
        self.handler = MakeHandler()
        self.parser = make_parser()
        self.parser.setContentHandler(self.handler)

        try:
            PS(xml_string, self.handler)
        except Exception as exc:
            print('ERROR while xml parsing')
            print(exc)


    def get_model(self):
        """
        As it says
        """
        return self.handler.model


class XmlException(Exception):
    """
    For exception identification
    """
    def __init__(self, mess):
        self.message = mess






















